import logging

from aprsd import packets, plugin

LOG = logging.getLogger("APRSD")


class HelloPlugin(plugin.APRSDRegexCommandPluginBase):
    """Hello World."""

    version = "1.0"
    # matches any string starting with h or H
    command_regex = "^[hH]"
    command_name = "hello"

    def process(self, packet: packets.MessagePacket):
        LOG.info("HelloPlugin")
        reply = f"Hello '{packet.from_call}'"
        return reply



import logging
import re
import time
from dataclasses import dataclass, field
from datetime import datetime

# Due to a failure in python 3.8
from typing import Any, List, Optional, Type, TypeVar, Union

from aprslib import util as aprslib_util
from dataclasses_json import (
    CatchAll,
    DataClassJsonMixin,
    Undefined,
    dataclass_json,
)
from loguru import logger

from aprsd.utils import counter

# For mypy to be happy
A = TypeVar('A', bound='DataClassJsonMixin')
Json = Union[dict, list, str, int, float, bool, None]

LOG = logging.getLogger()
LOGU = logger

PACKET_TYPE_BULLETIN = 'bulletin'
PACKET_TYPE_MESSAGE = 'message'
PACKET_TYPE_ACK = 'ack'
PACKET_TYPE_REJECT = 'reject'
PACKET_TYPE_MICE = 'mic-e'
PACKET_TYPE_WX = 'wx'
PACKET_TYPE_WEATHER = 'weather'
PACKET_TYPE_OBJECT = 'object'
PACKET_TYPE_UNKNOWN = 'unknown'
PACKET_TYPE_STATUS = 'status'
PACKET_TYPE_BEACON = 'beacon'
PACKET_TYPE_THIRDPARTY = 'thirdparty'
PACKET_TYPE_TELEMETRY = 'telemetry-message'
PACKET_TYPE_UNCOMPRESSED = 'uncompressed'

NO_DATE = datetime(1900, 10, 24)


def _init_timestamp():
    """Build a unix style timestamp integer"""
    return int(round(time.time()))


def _init_send_time():
    # We have to use a datetime here, or the json encoder
    # Fails on a NoneType.
    return NO_DATE


def _init_msgNo():  # noqa: N802
    """For some reason __post__init doesn't get called.

    So in order to initialize the msgNo field in the packet
    we use this workaround.
    """
    c = counter.PacketCounter()
    c.increment()
    return c.value


def _translate_fields(raw: dict) -> dict:
    # Direct key checks instead of iteration
    if 'from' in raw:
        raw['from_call'] = raw.pop('from')
    if 'to' in raw:
        raw['to_call'] = raw.pop('to')

    # addresse overrides to_call
    if 'addresse' in raw:
        raw['to_call'] = raw['addresse']

    return raw


@dataclass_json
@dataclass(unsafe_hash=True)
class Packet:
    _type: str = field(default='Packet', hash=False)
    from_call: Optional[str] = field(default=None)
    to_call: Optional[str] = field(default=None)
    addresse: Optional[str] = field(default=None)
    format: Optional[str] = field(default=None)
    msgNo: Optional[str] = field(default=None)  # noqa: N815
    ackMsgNo: Optional[str] = field(default=None)  # noqa: N815
    packet_type: Optional[str] = field(default=None)
    timestamp: float = field(default_factory=_init_timestamp, compare=False, hash=False)
    # Holds the raw text string to be sent over the wire
    # or holds the raw string from input packet
    raw: Optional[str] = field(default=None, compare=False, hash=False)
    raw_dict: dict = field(
        repr=False, default_factory=lambda: {}, compare=False, hash=False
    )
    # Built by calling prepare().  raw needs this built first.
    payload: Optional[str] = field(default=None)

    # Fields related to sending packets out
    send_count: int = field(repr=False, default=0, compare=False, hash=False)
    retry_count: int = field(repr=False, default=3, compare=False, hash=False)
    last_send_time: float = field(repr=False, default=0, compare=False, hash=False)
    # Was the packet acked?
    acked: bool = field(repr=False, default=False, compare=False, hash=False)
    # Was the packet previously processed (for dupe checking)
    processed: bool = field(repr=False, default=False, compare=False, hash=False)

    # Do we allow this packet to be saved to send later?
    allow_delay: bool = field(repr=False, default=True, compare=False, hash=False)
    path: List[str] = field(default_factory=list, compare=False, hash=False)
    via: Optional[str] = field(default=None, compare=False, hash=False)

    def get(self, key: str, default: Optional[str] = None):
        return getattr(self, key, default)

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        return f'{self.from_call}:{self.addresse}:{self.msgNo}'

    def update_timestamp(self) -> None:
        self.timestamp = _init_timestamp()

    @property
    def human_info(self) -> str:
        """Build a human readable string for this packet.

        This doesn't include the from to and type, but just
        the human readable payload.
        """
        self.prepare()
        msg = self._filter_for_send(self.raw).rstrip('\n')
        return msg

    def prepare(self, create_msg_number=False) -> None:
        """Do stuff here that is needed prior to sending over the air."""
        # now build the raw message for sending
        if not self.msgNo and create_msg_number:
            self.msgNo = _init_msgNo()
        self._build_payload()
        self._build_raw()

    def _build_payload(self) -> None:
        """The payload is the non headers portion of the packet."""
        if not self.to_call:
            raise ValueError(
                "to_call isn't set. Must set to_call before calling prepare()"
            )

        # The base packet class has no real payload
        self.payload = f':{self.to_call.ljust(9)}'

    def _build_raw(self) -> None:
        """Build the self.raw which is what is sent over the air."""
        self.raw = '{}>APZ100:{}'.format(
            self.from_call,
            self.payload,
        )

    def _filter_for_send(self, msg) -> str:
        """Filter and format message string for FCC."""
        # max?  ftm400 displays 64, raw msg shows 74
        # and ftm400-send is max 64.  setting this to
        # 67 displays 64 on the ftm400. (+3 {01 suffix)
        # feature req: break long ones into two msgs
        if not msg:
            return ''

        message = msg[:67]
        # We all miss George Carlin
        return re.sub(
            'fuck|shit|cunt|piss|cock|bitch',
            '****',
            message,
            flags=re.IGNORECASE,
        )

    def __str__(self) -> str:
        """Show the raw version of the packet"""
        self.prepare()
        if not self.raw:
            raise ValueError('self.raw is unset')
        return self.raw

    def __repr__(self) -> str:
        """Build the repr version of the packet."""
        return (
            f'{self.__class__.__name__}: From: {self.from_call}     To: {self.to_call}'
        )


@dataclass_json
@dataclass(unsafe_hash=True)
class AckPacket(Packet):
    _type: str = field(default='AckPacket', hash=False)

    def _build_payload(self):
        self.payload = f':{self.to_call: <9}:ack{self.msgNo}'


@dataclass_json
@dataclass(unsafe_hash=True)
class BulletinPacket(Packet):
    _type: str = 'BulletinPacket'
    # Holds the encapsulated packet
    bid: Optional[str] = field(default='1')
    message_text: Optional[str] = field(default=None)

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        return f'{self.from_call}:BLN{self.bid}'

    @property
    def human_info(self) -> str:
        return f'BLN{self.bid} {self.message_text}'

    def _build_payload(self) -> None:
        self.payload = f':BLN{self.bid:<9}:{self.message_text}'


@dataclass_json
@dataclass(unsafe_hash=True)
class RejectPacket(Packet):
    _type: str = field(default='RejectPacket', hash=False)
    response: Optional[str] = field(default=None)

    def __post__init__(self):
        if self.response:
            LOG.warning('Response set!')

    def _build_payload(self):
        self.payload = f':{self.to_call: <9}:rej{self.msgNo}'


@dataclass_json
@dataclass(unsafe_hash=True)
class MessagePacket(Packet):
    _type: str = field(default='MessagePacket', hash=False)
    message_text: Optional[str] = field(default=None)

    @property
    def human_info(self) -> str:
        self.prepare()
        return self._filter_for_send(self.message_text).rstrip('\n')

    def _build_payload(self):
        if self.msgNo:
            self.payload = ':{}:{}{{{}'.format(
                self.to_call.ljust(9),
                self._filter_for_send(self.message_text).rstrip('\n'),
                str(self.msgNo),
            )
        else:
            self.payload = ':{}:{}'.format(
                self.to_call.ljust(9),
                self._filter_for_send(self.message_text).rstrip('\n'),
            )


@dataclass_json
@dataclass(unsafe_hash=True)
class StatusPacket(Packet):
    _type: str = field(default='StatusPacket', hash=False)
    status: Optional[str] = field(default=None)
    messagecapable: bool = field(default=False)
    comment: Optional[str] = field(default=None)
    raw_timestamp: Optional[str] = field(default=None)

    def _build_payload(self):
        self.payload = ':{}:{}{{{}'.format(
            self.to_call.ljust(9),
            self._filter_for_send(self.status).rstrip('\n'),
            str(self.msgNo),
        )

    @property
    def human_info(self) -> str:
        self.prepare()
        return self.status


@dataclass_json
@dataclass(unsafe_hash=True)
class GPSPacket(Packet):
    _type: str = field(default='GPSPacket', hash=False)
    latitude: float = field(default=0.00)
    longitude: float = field(default=0.00)
    altitude: float = field(default=0.00)
    rng: float = field(default=0.00)
    posambiguity: int = field(default=0)
    messagecapable: bool = field(default=False)
    comment: Optional[str] = field(default=None)
    symbol: str = field(default='l')
    symbol_table: str = field(default='/')
    raw_timestamp: Optional[str] = field(default=None)
    object_name: Optional[str] = field(default=None)
    object_format: Optional[str] = field(default=None)
    alive: Optional[bool] = field(default=None)
    course: Optional[int] = field(default=None)
    speed: Optional[float] = field(default=None)
    phg: Optional[str] = field(default=None)
    phg_power: Optional[int] = field(default=None)
    phg_height: Optional[float] = field(default=None)
    phg_gain: Optional[int] = field(default=None)
    phg_dir: Optional[str] = field(default=None)
    phg_range: Optional[float] = field(default=None)
    phg_rate: Optional[int] = field(default=None)
    # http://www.aprs.org/datum.txt
    daodatumbyte: Optional[str] = field(default=None)

    def _build_time_zulu(self):
        """Build the timestamp in UTC/zulu."""
        if self.timestamp:
            return datetime.utcfromtimestamp(self.timestamp).strftime('%d%H%M')

    def _build_payload(self):
        """The payload is the non headers portion of the packet."""
        time_zulu = self._build_time_zulu()
        lat = aprslib_util.latitude_to_ddm(self.latitude)
        long = aprslib_util.longitude_to_ddm(self.longitude)
        payload = [
            '@' if self.timestamp else '!',
            time_zulu,
            lat,
            self.symbol_table,
            long,
            self.symbol,
        ]

        if self.comment:
            payload.append(self._filter_for_send(self.comment))

        self.payload = ''.join(payload)

    def _build_raw(self):
        self.raw = f'{self.from_call}>{self.to_call},WIDE2-1:{self.payload}'

    @property
    def human_info(self) -> str:
        h_str = []
        h_str.append(f'Lat:{self.latitude:03.3f}')
        h_str.append(f'Lon:{self.longitude:03.3f}')
        if self.altitude:
            h_str.append(f'Altitude {self.altitude:03.0f}')
        if self.speed:
            h_str.append(f'Speed {self.speed:03.0f}MPH')
        if self.course:
            h_str.append(f'Course {self.course:03.0f}')
        if self.rng:
            h_str.append(f'RNG {self.rng:03.0f}')
        if self.phg:
            h_str.append(f'PHG {self.phg}')

        return ' '.join(h_str)


@dataclass_json
@dataclass(unsafe_hash=True)
class BeaconPacket(GPSPacket):
    _type: str = field(default='BeaconPacket', hash=False)

    def _build_payload(self):
        """The payload is the non headers portion of the packet."""
        time_zulu = self._build_time_zulu()
        lat = aprslib_util.latitude_to_ddm(self.latitude)
        lon = aprslib_util.longitude_to_ddm(self.longitude)

        self.payload = f'@{time_zulu}z{lat}{self.symbol_table}{lon}'

        if self.comment:
            comment = self._filter_for_send(self.comment)
            self.payload = f'{self.payload}{self.symbol}{comment}'
        else:
            self.payload = f'{self.payload}{self.symbol}APRSD Beacon'

    def _build_raw(self):
        self.raw = f'{self.from_call}>APZ100:{self.payload}'

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        if self.raw_timestamp:
            return f'{self.from_call}:{self.raw_timestamp}'
        else:
            return f'{self.from_call}:{self.human_info.replace(" ", "")}'

    @property
    def human_info(self) -> str:
        h_str = []
        h_str.append(f'Lat:{self.latitude:03.3f}')
        h_str.append(f'Lon:{self.longitude:03.3f}')
        h_str.append(f'{self.comment}')
        return ' '.join(h_str)


@dataclass_json
@dataclass(unsafe_hash=True)
class MicEPacket(GPSPacket):
    _type: str = field(default='MicEPacket', hash=False)
    messagecapable: bool = False
    mbits: Optional[str] = None
    mtype: Optional[str] = None
    telemetry: Optional[dict] = field(default=None, hash=False)
    # in MPH
    speed: float = 0.00
    # 0 to 360
    course: int = 0

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        return f'{self.from_call}:{self.human_info.replace(" ", "")}'

    @property
    def human_info(self) -> str:
        h_info = super().human_info
        return f'{h_info} {self.mbits} mbits'


@dataclass_json
@dataclass(unsafe_hash=True)
class TelemetryPacket(GPSPacket):
    _type: str = field(default='TelemetryPacket', hash=False)
    messagecapable: bool = False
    mbits: Optional[str] = None
    mtype: Optional[str] = None
    telemetry: Optional[dict] = field(default=None)
    tPARM: Optional[list[str]] = field(default=None, hash=False)  # noqa: N815
    tUNIT: Optional[list[str]] = field(default=None, hash=False)  # noqa: N815
    # in MPH
    speed: float = 0.00
    # 0 to 360
    course: int = 0

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        if self.raw_timestamp:
            return f'{self.from_call}:{self.raw_timestamp}'
        else:
            return f'{self.from_call}:{self.human_info.replace(" ", "")}'

    @property
    def human_info(self) -> str:
        h_info = super().human_info
        return f'{h_info} {self.telemetry}'


@dataclass_json
@dataclass(unsafe_hash=True)
class ObjectPacket(GPSPacket):
    _type: str = field(default='ObjectPacket', hash=False)
    alive: bool = True
    raw_timestamp: Optional[str] = None
    symbol: str = field(default='r')
    # in MPH
    speed: float = 0.00
    # 0 to 360
    course: int = 0

    def _build_payload(self):
        time_zulu = self._build_time_zulu()
        lat = aprslib_util.latitude_to_ddm(self.latitude)
        long = aprslib_util.longitude_to_ddm(self.longitude)

        self.payload = f'*{time_zulu}z{lat}{self.symbol_table}{long}{self.symbol}'

        if self.comment:
            comment = self._filter_for_send(self.comment)
            self.payload = f'{self.payload}{comment}'

    def _build_raw(self):
        """
        REPEAT builds packets like
        reply = "{}>APZ100:;{:9s}*{}z{}r{:.3f}MHz {} {}".format(
                fromcall, callsign, time_zulu, latlon, freq, uplink_tone, offset,
            )
        where fromcall is the callsign that is sending the packet
        callsign is the station callsign for the object
        The frequency, uplink_tone, offset is part of the comment
        """

        self.raw = f'{self.from_call}>APZ100:;{self.to_call:9s}{self.payload}'

    @property
    def human_info(self) -> str:
        h_info = super().human_info
        return f'{h_info} {self.comment}'


@dataclass(unsafe_hash=True)
class WeatherPacket(GPSPacket, DataClassJsonMixin):
    _type: str = field(default='WeatherPacket', hash=False)
    symbol: str = '_'
    wind_speed: float = 0.00
    wind_direction: int = 0
    wind_gust: float = 0.00
    temperature: float = 0.00
    # in inches.  1.04 means 1.04 inches
    rain_1h: float = 0.00
    rain_24h: float = 0.00
    rain_since_midnight: float = 0.00
    humidity: int = 0
    pressure: float = 0.00
    comment: Optional[str] = field(default=None)
    luminosity: Optional[int] = field(default=None)
    wx_raw_timestamp: Optional[str] = field(default=None)
    course: Optional[int] = field(default=None)
    speed: Optional[float] = field(default=None)

    def _translate(self, raw: dict) -> dict:
        for key in raw['weather']:
            raw[key] = raw['weather'][key]

        # If we have the broken aprslib, then we need to
        # Convert the course and speed to wind_speed and wind_direction
        # aprslib issue #80
        # https://github.com/rossengeorgiev/aprs-python/issues/80
        # Wind speed and course is option in the SPEC.
        # For some reason aprslib multiplies the speed by 1.852.
        if 'wind_speed' not in raw and 'wind_direction' not in raw:
            # Most likely this is the broken aprslib
            # So we need to convert the wind_gust speed
            raw['wind_gust'] = round(raw.get('wind_gust', 0) / 0.44704, 3)
        if 'wind_speed' not in raw:
            wind_speed = raw.get('speed')
            if wind_speed:
                raw['wind_speed'] = round(wind_speed / 1.852, 3)
                raw['weather']['wind_speed'] = raw['wind_speed']
            if 'speed' in raw:
                del raw['speed']
            # Let's adjust the rain numbers as well, since it's wrong
            raw['rain_1h'] = round((raw.get('rain_1h', 0) / 0.254) * 0.01, 3)
            raw['weather']['rain_1h'] = raw['rain_1h']
            raw['rain_24h'] = round((raw.get('rain_24h', 0) / 0.254) * 0.01, 3)
            raw['weather']['rain_24h'] = raw['rain_24h']
            raw['rain_since_midnight'] = round(
                (raw.get('rain_since_midnight', 0) / 0.254) * 0.01, 3
            )
            raw['weather']['rain_since_midnight'] = raw['rain_since_midnight']

        if 'wind_direction' not in raw:
            wind_direction = raw.get('course')
            if wind_direction:
                raw['wind_direction'] = wind_direction
                raw['weather']['wind_direction'] = raw['wind_direction']
            if 'course' in raw:
                del raw['course']

        del raw['weather']
        return raw

    @classmethod
    def from_dict(cls: Type[A], kvs: Json, *, infer_missing=False) -> A:
        """Create from a dictionary that has come directly from aprslib parse"""
        raw = cls._translate(cls, kvs)  # type: ignore
        return super().from_dict(raw)

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        if self.raw_timestamp:
            return f'{self.from_call}:{self.raw_timestamp}'
        elif self.wx_raw_timestamp:
            return f'{self.from_call}:{self.wx_raw_timestamp}'

    @property
    def human_info(self) -> str:
        h_str = []
        h_str.append(f'Temp {self.temperature:03.0f}F')
        h_str.append(f'Humidity {self.humidity}%')
        h_str.append(f'Wind {self.wind_speed:03.0f}MPH@{self.wind_direction}')
        h_str.append(f'Pressure {self.pressure}mb')
        h_str.append(f'Rain {self.rain_24h}in/24hr')

        return ' '.join(h_str)

    def _build_payload(self):
        """Build an uncompressed weather packet

         Format =

        _CSE/SPDgXXXtXXXrXXXpXXXPXXXhXXbXXXXX%type NEW FORMAT APRS793 June 97
                                                   NOT BACKWARD COMPATIBLE


         Where: CSE/SPD is wind direction and sustained 1 minute speed
         t is in degrees F

         r is Rain per last 60 minutes
             1.04 inches of rain will show as r104
         p is precipitation per last 24 hours (sliding 24 hour window)
         P is precip per last 24 hours since midnight
         b is Baro in tenths of a mb
         h is humidity in percent. 00=100
         g is Gust (peak winds in last 5 minutes)
         # is the raw rain counter for remote WX stations
         See notes on remotes below
         % shows software type d=Dos, m=Mac, w=Win, etc
         type shows type of WX instrument

        """
        time_zulu = self._build_time_zulu()

        contents = [
            f'@{time_zulu}z{self.latitude}{self.symbol_table}',
            f'{self.longitude}{self.symbol}',
            f'{self.wind_direction:03d}',
            # Speed = sustained 1 minute wind speed in mph
            f'{self.symbol_table}',
            f'{self.wind_speed:03.0f}',
            # wind gust (peak wind speed in mph in the last 5 minutes)
            f'g{self.wind_gust:03.0f}',
            # Temperature in degrees F
            f't{self.temperature:03.0f}',
            # Rainfall (in hundredths of an inch) in the last hour
            f'r{self.rain_1h * 100:03.0f}',
            # Rainfall (in hundredths of an inch) in last 24 hours
            f'p{self.rain_24h * 100:03.0f}',
            # Rainfall (in hundredths of an inch) since midnigt
            f'P{self.rain_since_midnight * 100:03.0f}',
            # Humidity
            f'h{self.humidity:02d}',
            # Barometric pressure (in tenths of millibars/tenths of hPascal)
            f'b{self.pressure:05.0f}',
        ]
        if self.comment:
            comment = self._filter_for_send(self.comment)
            contents.append(comment)
        self.payload = ''.join(contents)

    def _build_raw(self):
        self.raw = f'{self.from_call}>{self.to_call},WIDE1-1,WIDE2-1:{self.payload}'


@dataclass(unsafe_hash=True)
class ThirdPartyPacket(Packet, DataClassJsonMixin):
    _type: str = 'ThirdPartyPacket'
    # Holds the encapsulated packet
    subpacket: Optional[type[Packet]] = field(default=None, compare=True, hash=False)

    def __repr__(self):
        """Build the repr version of the packet."""
        repr_str = (
            f'{self.__class__.__name__}:'
            f' From: {self.from_call}  '
            f' To: {self.to_call}  '
            f' Subpacket: {repr(self.subpacket)}'
        )

        return repr_str

    @classmethod
    def from_dict(cls: Type[A], kvs: Json, *, infer_missing=False) -> A:
        obj = super().from_dict(kvs)
        obj.subpacket = factory(obj.subpacket)  # type: ignore
        return obj

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        return f'{self.from_call}:{self.subpacket.key}'

    @property
    def human_info(self) -> str:
        sub_info = self.subpacket.human_info
        return f'{self.from_call}->{self.to_call} {sub_info}'


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass(unsafe_hash=True)
class UnknownPacket:
    """Catchall Packet for things we don't know about.

    All of the unknown attributes are stored in the unknown_fields
    """

    unknown_fields: CatchAll
    _type: str = 'UnknownPacket'
    from_call: Optional[str] = field(default=None)
    to_call: Optional[str] = field(default=None)
    msgNo: str = field(default_factory=_init_msgNo)  # noqa: N815
    format: Optional[str] = field(default=None)
    timestamp: float = field(default_factory=_init_timestamp, compare=False, hash=False)
    raw: Optional[str] = field(default=None)
    raw_dict: dict = field(
        repr=False, default_factory=lambda: {}, compare=False, hash=False
    )
    path: List[str] = field(default_factory=list, compare=False, hash=False)
    packet_type: Optional[str] = field(default=None)
    via: Optional[str] = field(default=None, compare=False, hash=False)
    # Was the packet previously processed (for dupe checking)
    processed: bool = field(repr=False, default=False, compare=False, hash=False)

    @property
    def key(self) -> str:
        """Build a key for finding this packet in a dict."""
        return f'{self.from_call}:{self.packet_type}:{self.to_call}'

    @property
    def human_info(self) -> str:
        return str(self.unknown_fields)


TYPE_LOOKUP: dict[str, type[Packet]] = {
    PACKET_TYPE_BULLETIN: BulletinPacket,
    PACKET_TYPE_WX: WeatherPacket,
    PACKET_TYPE_WEATHER: WeatherPacket,
    PACKET_TYPE_MESSAGE: MessagePacket,
    PACKET_TYPE_ACK: AckPacket,
    PACKET_TYPE_REJECT: RejectPacket,
    PACKET_TYPE_MICE: MicEPacket,
    PACKET_TYPE_OBJECT: ObjectPacket,
    PACKET_TYPE_STATUS: StatusPacket,
    PACKET_TYPE_BEACON: BeaconPacket,
    PACKET_TYPE_UNKNOWN: UnknownPacket,
    PACKET_TYPE_THIRDPARTY: ThirdPartyPacket,
    PACKET_TYPE_TELEMETRY: TelemetryPacket,
}


def get_packet_type(packet: dict) -> str:
    """Decode the packet type from the packet."""

    pkt_format = packet.get('format')
    msg_response = packet.get('response')
    packet_type = PACKET_TYPE_UNKNOWN
    if pkt_format == 'message' and msg_response == 'ack':
        packet_type = PACKET_TYPE_ACK
    elif pkt_format == 'message' and msg_response == 'rej':
        packet_type = PACKET_TYPE_REJECT
    elif pkt_format == 'message':
        packet_type = PACKET_TYPE_MESSAGE
    elif pkt_format == 'mic-e':
        packet_type = PACKET_TYPE_MICE
    elif pkt_format == 'object':
        packet_type = PACKET_TYPE_OBJECT
    elif pkt_format == 'status':
        packet_type = PACKET_TYPE_STATUS
    elif pkt_format == PACKET_TYPE_BULLETIN:
        packet_type = PACKET_TYPE_BULLETIN
    elif pkt_format == PACKET_TYPE_BEACON:
        packet_type = PACKET_TYPE_BEACON
    elif pkt_format == PACKET_TYPE_TELEMETRY:
        packet_type = PACKET_TYPE_TELEMETRY
    elif pkt_format == PACKET_TYPE_WX:
        packet_type = PACKET_TYPE_WEATHER
    elif pkt_format == PACKET_TYPE_UNCOMPRESSED:
        if packet.get('symbol') == '_':
            packet_type = PACKET_TYPE_WEATHER
    elif pkt_format == PACKET_TYPE_THIRDPARTY:
        packet_type = PACKET_TYPE_THIRDPARTY

    if packet_type == PACKET_TYPE_UNKNOWN:
        if 'latitude' in packet:
            packet_type = PACKET_TYPE_BEACON
        else:
            packet_type = PACKET_TYPE_UNKNOWN
    return packet_type


def is_message_packet(packet: dict) -> bool:
    return get_packet_type(packet) == PACKET_TYPE_MESSAGE


def is_ack_packet(packet: dict) -> bool:
    return get_packet_type(packet) == PACKET_TYPE_ACK


def is_mice_packet(packet: dict[Any, Any]) -> bool:
    return get_packet_type(packet) == PACKET_TYPE_MICE


def factory(raw_packet: dict[Any, Any]) -> type[Packet]:
    """Factory method to create a packet from a raw packet string."""
    raw = raw_packet
    if '_type' in raw:
        cls = globals()[raw['_type']]
        return cls.from_dict(raw)

    raw['raw_dict'] = raw.copy()
    raw = _translate_fields(raw)

    packet_type = get_packet_type(raw)

    raw['packet_type'] = packet_type
    packet_class = TYPE_LOOKUP[packet_type]
    if packet_type == PACKET_TYPE_WX:
        # the weather information is in a dict
        # this brings those values out to the outer dict
        packet_class = WeatherPacket
    elif packet_type == PACKET_TYPE_OBJECT and 'weather' in raw:
        packet_class = WeatherPacket
    elif packet_type == PACKET_TYPE_UNKNOWN:
        # Try and figure it out here
        if 'latitude' in raw:
            packet_class = GPSPacket
        else:
            # LOG.warning(raw)
            packet_class = UnknownPacket

    raw.get('addresse', raw.get('to_call'))

    # TODO: Find a global way to enable/disable this
    # LOGU.opt(colors=True).info(
    #     f"factory(<green>{packet_type: <8}</green>):"
    #     f"(<red>{packet_class.__name__: <13}</red>): "
    #     f"<light-blue>{raw.get('from_call'): <9}</light-blue> -> <cyan>{to: <9}</cyan>")
    # LOG.info(raw.get('msgNo'))

    return packet_class().from_dict(raw)  # type: ignore



import logging
from typing import Optional

from haversine import Unit, haversine
from loguru import logger
from oslo_config import cfg

from aprsd import utils
from aprsd.packets.core import AckPacket, GPSPacket, RejectPacket

LOG = logging.getLogger()
LOGU = logger
CONF = cfg.CONF

FROM_COLOR = 'fg #C70039'
TO_COLOR = 'fg #D033FF'
TX_COLOR = 'red'
RX_COLOR = 'green'
PACKET_COLOR = 'cyan'
DISTANCE_COLOR = 'fg #FF5733'
DEGREES_COLOR = 'fg #FFA900'


def log_multiline(
    packet, tx: Optional[bool] = False, header: Optional[bool] = True
) -> None:
    """LOG a packet to the logfile."""
    if not CONF.enable_packet_logging:
        return
    if CONF.log_packet_format == 'compact':
        return

    # asdict(packet)
    logit = ['\n']
    name = packet.__class__.__name__

    if isinstance(packet, AckPacket):
        pkt_max_send_count = CONF.default_ack_send_count
    else:
        pkt_max_send_count = CONF.default_packet_send_count

    if header:
        if tx:
            header_str = f'<{TX_COLOR}>TX</{TX_COLOR}>'
            logit.append(
                f'{header_str}________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>  '
                f'TX:{packet.send_count + 1} of {pkt_max_send_count}',
            )
        else:
            header_str = f'<{RX_COLOR}>RX</{RX_COLOR}>'
            logit.append(
                f'{header_str}________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>)',
            )

    else:
        header_str = ''
        logit.append(f'__________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>)')
    # log_list.append(f"  Packet  : {packet.__class__.__name__}")
    if packet.msgNo:
        logit.append(f'  Msg #   : {packet.msgNo}')
    if packet.from_call:
        logit.append(f'  From    : <{FROM_COLOR}>{packet.from_call}</{FROM_COLOR}>')
    if packet.to_call:
        logit.append(f'  To      : <{TO_COLOR}>{packet.to_call}</{TO_COLOR}>')
    if hasattr(packet, 'path') and packet.path:
        logit.append(f'  Path    : {"=>".join(packet.path)}')
    if hasattr(packet, 'via') and packet.via:
        logit.append(f'  VIA     : {packet.via}')

    if not isinstance(packet, AckPacket) and not isinstance(packet, RejectPacket):
        msg = packet.human_info

        if msg:
            msg = msg.replace('<', '\\<')
            logit.append(f'  Info    : <light-yellow><b>{msg}</b></light-yellow>')

    if hasattr(packet, 'comment') and packet.comment:
        logit.append(f'  Comment : {packet.comment}')

    raw = packet.raw.replace('<', '\\<')
    logit.append(f'  Raw     : <fg #828282>{raw}</fg #828282>')
    logit.append(f'{header_str}________(<{PACKET_COLOR}>{name}</{PACKET_COLOR}>)')

    LOGU.opt(colors=True).info('\n'.join(logit))
    LOG.debug(repr(packet))


def log(
    packet,
    tx: Optional[bool] = False,
    header: Optional[bool] = True,
    packet_count: Optional[int] = None,
) -> None:
    if not CONF.enable_packet_logging:
        return
    if CONF.log_packet_format == 'multiline':
        log_multiline(packet, tx, header)
        return

    if not packet_count:
        packet_count = ''
    else:
        packet_count = f'({packet_count:d})'

    logit = []
    name = packet.__class__.__name__
    if isinstance(packet, AckPacket):
        pkt_max_send_count = CONF.default_ack_send_count
    else:
        pkt_max_send_count = CONF.default_packet_send_count

    if header:
        if tx:
            via_color = 'red'
            arrow = f'<{via_color}>\u2192</{via_color}>'
            logit.append(
                f'<red>TX{packet_count}\u2191</red> '
                f'<cyan>{name}</cyan>'
                f':{packet.msgNo}'
                f' ({packet.send_count + 1} of {pkt_max_send_count})',
            )
        else:
            via_color = 'fg #1AA730'
            arrow = f'<{via_color}>\u2192</{via_color}>'
            f'<{via_color}><-</{via_color}>'
            logit.append(
                f'<fg #1AA730>RX{packet_count}\u2193</fg #1AA730> '
                f'<cyan>{name}</cyan>'
                f':{packet.msgNo}',
            )
    else:
        via_color = 'green'
        arrow = f'<{via_color}>-></{via_color}>'
        logit.append(
            f'<cyan>{name}</cyan>:{packet.msgNo}',
        )

    tmp = None
    if packet.path:
        tmp = f'{arrow}'.join(packet.path) + f'{arrow} '

    logit.append(
        f'<{FROM_COLOR}>{packet.from_call}</{FROM_COLOR}> {arrow}'
        f'{tmp if tmp else " "}'
        f'<{TO_COLOR}>{packet.to_call}</{TO_COLOR}>',
    )

    if not isinstance(packet, AckPacket) and not isinstance(packet, RejectPacket):
        logit.append(':')
        msg = packet.human_info

        if msg:
            msg = msg.replace('<', '\\<')
            logit.append(f'<light-yellow><b>{msg}</b></light-yellow>')

    # is there distance information?
    if isinstance(packet, GPSPacket) and CONF.latitude and CONF.longitude:
        my_coords = (float(CONF.latitude), float(CONF.longitude))
        packet_coords = (float(packet.latitude), float(packet.longitude))
        try:
            bearing = utils.calculate_initial_compass_bearing(my_coords, packet_coords)
        except Exception as e:
            LOG.error(f'Failed to calculate bearing: {e}')
            bearing = 0
        logit.append(
            f' : <{DEGREES_COLOR}>{utils.degrees_to_cardinal(bearing, full_string=True)}</{DEGREES_COLOR}>'
            f'<{DISTANCE_COLOR}>@{haversine(my_coords, packet_coords, unit=Unit.MILES):.2f}miles</{DISTANCE_COLOR}>',
        )

    LOGU.opt(colors=True).info(' '.join(logit))
    log_multiline(packet, tx, header)



from aprsd import plugin, threads
from aprsd.packets import core

FAKE_MESSAGE_TEXT = "fake MeSSage"
FAKE_FROM_CALLSIGN = "KFAKE"
FAKE_TO_CALLSIGN = "KMINE"


def fake_packet(
    fromcall=FAKE_FROM_CALLSIGN,
    tocall=FAKE_TO_CALLSIGN,
    message=None,
    msg_number=None,
    message_format=core.PACKET_TYPE_MESSAGE,
    response=None,
):
    packet_dict = {
        "from": fromcall,
        "addresse": tocall,
        "to": tocall,
        "format": message_format,
        "raw": "",
    }
    if message:
        packet_dict["message_text"] = message

    if msg_number:
        packet_dict["msgNo"] = str(msg_number)

    if response:
        packet_dict["response"] = response

    return core.factory(packet_dict)


def fake_ack_packet():
    return fake_packet(
        msg_number=12,
        response=core.PACKET_TYPE_ACK,
    )


class FakeBaseNoThreadsPlugin(plugin.APRSDPluginBase):
    version = "1.0"

    def setup(self):
        self.enabled = True

    def filter(self, packet):
        return None

    def process(self, packet):
        return "process"


class FakeThread(threads.APRSDThread):
    def __init__(self):
        super().__init__("FakeThread")

    def loop(self):
        return False


class FakeBaseThreadsPlugin(plugin.APRSDPluginBase):
    version = "1.0"

    def setup(self):
        self.enabled = True

    def filter(self, packet):
        return None

    def process(self, packet):
        return "process"

    def create_threads(self):
        return FakeThread()


class FakeRegexCommandPlugin(plugin.APRSDRegexCommandPluginBase):
    version = "1.0"
    command_regex = "^[fF]"
    command_name = "fake"

    def process(self, packet):
        return FAKE_MESSAGE_TEXT


class FakeWatchListPlugin(plugin.APRSDWatchListPluginBase):
    def process(self, packet):
        return FAKE_MESSAGE_TEXT



import logging
import sys
import time

import aprslib
import click
from aprslib.exceptions import LoginError
from oslo_config import cfg

import aprsd
import aprsd.packets  # noqa : F401
from aprsd import (
    cli_helper,
    conf,  # noqa : F401
    packets,
    utils,
)
from aprsd.client.client import APRSDClient
from aprsd.main import cli
from aprsd.packets import collector
from aprsd.packets import log as packet_log
from aprsd.threads import tx

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    '--aprs-login',
    envvar='APRS_LOGIN',
    show_envvar=True,
    help='What callsign to send the message from. Defaults to config entry.',
)
@click.option(
    '--aprs-password',
    envvar='APRS_PASSWORD',
    show_envvar=True,
    help='the APRS-IS password for APRS_LOGIN. Defaults to config entry.',
)
@click.option(
    '--no-ack',
    '-n',
    is_flag=True,
    show_default=True,
    default=False,
    help="Don't wait for an ack, just sent it to APRS-IS and bail.",
)
@click.option(
    '--wait-response',
    '-w',
    is_flag=True,
    show_default=True,
    default=False,
    help='Wait for a response to the message?',
)
@click.option('--raw', default=None, help='Send a raw message.  Implies --no-ack')
@click.argument('tocallsign', required=True)
@click.argument('command', nargs=-1, required=True)
@click.pass_context
@cli_helper.process_standard_options
def send_message(
    ctx,
    aprs_login,
    aprs_password,
    no_ack,
    wait_response,
    raw,
    tocallsign,
    command,
):
    """Send a message to a callsign via APRS_IS."""
    global got_ack, got_response
    quiet = ctx.obj['quiet']

    if not aprs_login:
        if CONF.aprs_network.login == conf.client.DEFAULT_LOGIN:
            click.echo('Must set --aprs_login or APRS_LOGIN')
            ctx.exit(-1)
            return
        else:
            aprs_login = CONF.aprs_network.login

    if not aprs_password:
        if not CONF.aprs_network.password:
            click.echo('Must set --aprs-password or APRS_PASSWORD')
            ctx.exit(-1)
            return
        else:
            aprs_password = CONF.aprs_network.password

    LOG.info(f'Python version: {sys.version}')
    LOG.info(f'APRSD SEND_MESSAGE Started version: {aprsd.__version__}')
    utils.package.log_installed_extensions_and_plugins()
    if type(command) is tuple:
        command = ' '.join(command)
    if not quiet:
        if raw:
            LOG.info(f"L'{aprs_login}' R'{raw}'")
        else:
            LOG.info(f"L'{aprs_login}' To'{tocallsign}' C'{command}'")

    got_ack = False
    got_response = False

    def rx_packet(packet):
        global got_ack, got_response
        cl = APRSDClient()
        packet = cl.decode_packet(packet)
        collector.PacketCollector().rx(packet)
        packet_log.log(packet, tx=False)
        # LOG.debug("Got packet back {}".format(packet))
        if isinstance(packet, packets.AckPacket):
            got_ack = True
        else:
            got_response = True
            from_call = packet.from_call
            our_call = CONF.callsign.lower()
            tx.send(
                packets.AckPacket(
                    from_call=our_call,
                    to_call=from_call,
                    msgNo=packet.msgNo,
                ),
                direct=True,
            )

        if got_ack:
            if wait_response:
                if got_response:
                    sys.exit(0)
            else:
                sys.exit(0)

    try:
        APRSDClient().client  # noqa: B018
    except LoginError:
        sys.exit(-1)

    # Send a message
    # then we setup a consumer to rx messages
    # We should get an ack back as well as a new message
    # we should bail after we get the ack and send an ack back for the
    # message
    if raw:
        tx.send(
            packets.Packet(from_call='', to_call='', raw=raw),
            direct=True,
        )
        sys.exit(0)
    else:
        tx.send(
            packets.MessagePacket(
                from_call=aprs_login,
                to_call=tocallsign,
                message_text=command,
            ),
            direct=True,
        )

    if no_ack:
        sys.exit(0)

    try:
        # This will register a packet consumer with aprslib
        # When new packets come in the consumer will process
        # the packet
        aprs_client = APRSDClient()
        aprs_client.consumer(rx_packet, raw=False)
    except aprslib.exceptions.ConnectionDrop:
        LOG.error('Connection dropped, reconnecting')
        time.sleep(5)
        # Force the deletion of the client object connected to aprs
        # This will cause a reconnect, next time client.get_client()
        # is called
        aprs_client.reset()



import datetime
import logging

from oslo_config import cfg

from aprsd.packets import core
from aprsd.utils import objectstore


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class SeenList(objectstore.ObjectStoreMixin):
    """Global callsign seen list."""

    _instance = None
    data: dict = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.data = {}
        return cls._instance

    def stats(self, serializable=False):
        """Return the stats for the PacketTrack class."""
        with self.lock:
            return self.data

    def rx(self, packet: type[core.Packet]):
        """When we get a packet from the network, update the seen list."""
        with self.lock:
            callsign = None
            if packet.from_call:
                callsign = packet.from_call
            else:
                LOG.warning(f"Can't find FROM in packet {packet}")
                return
            if callsign not in self.data:
                self.data[callsign] = {
                    "last": None,
                    "count": 0,
                }
            self.data[callsign]["last"] = datetime.datetime.now()
            self.data[callsign]["count"] += 1

    def tx(self, packet: type[core.Packet]):
        """We don't care about TX packets."""



import random
import threading

import wrapt


MAX_PACKET_ID = 9999


class PacketCounter:
    """
    Global Packet ID counter class.

    This is a singleton-based class that keeps
    an incrementing counter for all packets to
    be sent. All new Packet objects get a new
    message ID, which is the next number available
    from the PacketCounter.

    """

    _instance = None
    lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        """Make this a singleton class."""
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._val = random.randint(1, MAX_PACKET_ID)  # Initialize counter
        return cls._instance

    @wrapt.synchronized(lock)
    def increment(self):
        """Increment the counter, reset if it exceeds MAX_PACKET_ID."""
        if self._val == MAX_PACKET_ID:
            self._val = 1
        else:
            self._val += 1

    @property
    @wrapt.synchronized(lock)
    def value(self):
        """Get the current value as a string."""
        return str(self._val)

    @wrapt.synchronized(lock)
    def __repr__(self):
        """String representation of the current value."""
        return str(self._val)

    @wrapt.synchronized(lock)
    def __str__(self):
        """String representation of the current value."""
        return str(self._val)



"""Utilities and helper functions."""

import errno
import functools
import math
import os
import re
import sys
import traceback

import update_checker

import aprsd

from .fuzzyclock import fuzzy  # noqa: F401

# Make these available by anyone importing
# aprsd.utils
from .ring_buffer import RingBuffer  # noqa: F401

if sys.version_info.major == 3 and sys.version_info.minor >= 3:
    from collections.abc import MutableMapping
else:
    from collections.abc import MutableMapping


def singleton(cls):
    """Make a class a Singleton class (only one instance)"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if wrapper_singleton.instance is None:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton


def env(*vars, **kwargs):
    """This returns the first environment variable set.
    if none are non-empty, defaults to '' or keyword arg default
    """
    for v in vars:
        value = os.environ.get(v, None)
        if value:
            return value
    return kwargs.get("default", "")


def mkdir_p(path):
    """Make directory and have it work in py2 and py3."""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >= 2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def insert_str(string, str_to_insert, index):
    return string[:index] + str_to_insert + string[index:]


def end_substr(original, substr):
    """Get the index of the end of the <substr>.

    So you can insert a string after <substr>
    """
    idx = original.find(substr)
    if idx != -1:
        idx += len(substr)
    return idx


def rgb_from_name(name):
    """Create an rgb tuple from a string."""
    hash = 0
    for char in name:
        hash = ord(char) + ((hash << 5) - hash)
    red = hash & 255
    green = (hash >> 8) & 255
    blue = (hash >> 16) & 255
    return red, green, blue


def hextriplet(colortuple):
    """Convert a color tuple to a hex triplet."""
    return "#" + "".join(f"{i:02X}" for i in colortuple)


def hex_from_name(name):
    """Create a hex color from a string."""
    return hextriplet(rgb_from_name(name))


def human_size(bytes, units=None):
    """Returns a human readable string representation of bytes"""
    if not units:
        units = [" bytes", "KB", "MB", "GB", "TB", "PB", "EB"]
    return str(bytes) + units[0] if bytes < 1024 else human_size(bytes >> 10, units[1:])


def strfdelta(tdelta, fmt="{hours:{width}}:{minutes:{width}}:{seconds:{width}}"):
    d = {
        "days": tdelta.days,
        "width": "02",
    }
    if tdelta.days > 0:
        fmt = "{days} days " + fmt

    d["hours"], rem = divmod(tdelta.seconds, 3600)
    d["minutes"], d["seconds"] = divmod(rem, 60)
    return fmt.format(**d)


def _check_version():
    # check for a newer version
    try:
        check = update_checker.UpdateChecker()
        result = check.check("aprsd", aprsd.__version__)
        if result:
            # Looks like there is an updated version.
            return 1, result
        else:
            return 0, "APRSD is up to date"
    except Exception:
        # probably can't get in touch with pypi for some reason
        # Lets put up an error and move on.  We might not
        # have internet in this aprsd deployment.
        return 1, "Couldn't check for new version of APRSD"


def flatten_dict(d, parent_key="", sep="."):
    """Flatten a dict to key.key.key = value."""
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def parse_delta_str(s):
    if "day" in s:
        m = re.match(
            r"(?P<days>[-\d]+) day[s]*, (?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d[\.\d+]*)",
            s,
        )
    else:
        m = re.match(r"(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d[\.\d+]*)", s)

    if m:
        return {key: float(val) for key, val in m.groupdict().items()}
    else:
        return {}


def load_entry_points(group):
    """Load all extensions registered to the given entry point group"""
    try:
        import importlib_metadata
    except ImportError:
        # For python 3.10 and later
        import importlib.metadata as importlib_metadata

    eps = importlib_metadata.entry_points(group=group)
    for ep in eps:
        try:
            ep.load()
        except Exception as e:
            print(
                f"Extension {ep.name} of group {group} failed to load with {e}",
                file=sys.stderr,
            )
            print(traceback.format_exc(), file=sys.stderr)


def calculate_initial_compass_bearing(point_a, point_b):
    """
    Calculates the bearing between two points.
    The formulae used is the following:
        θ = atan2(sin(Δlong).cos(lat2),
                  cos(lat1).sin(lat2) − sin(lat1).cos(lat2).cos(Δlong))
    :Parameters:
      - `pointA: The tuple representing the latitude/longitude for the
        first point. Latitude and longitude must be in decimal degrees
      - `pointB: The tuple representing the latitude/longitude for the
        second point. Latitude and longitude must be in decimal degrees
    :Returns:
      The bearing in degrees
    :Returns Type:
      float
    """
    if (type(point_a) != tuple) or (type(point_b) != tuple):  # noqa: E721
        raise TypeError("Only tuples are supported as arguments")

    lat1 = math.radians(float(point_a[0]))
    lat2 = math.radians(float(point_b[0]))

    diff_long = math.radians(float(point_b[1]) - float(point_a[1]))

    x = math.sin(diff_long) * math.cos(lat2)
    y = math.cos(lat1) * math.sin(lat2) - (
        math.sin(lat1) * math.cos(lat2) * math.cos(diff_long)
    )

    initial_bearing = math.atan2(x, y)

    # Now we have the initial bearing but math.atan2 return values
    # from -180° to + 180° which is not what we want for a compass bearing
    # The solution is to normalize the initial bearing as shown below
    initial_bearing = math.degrees(initial_bearing)
    compass_bearing = (initial_bearing + 360) % 360

    return compass_bearing


def degrees_to_cardinal(bearing, full_string=False):
    if full_string:
        directions = [
            "North",
            "North-Northeast",
            "Northeast",
            "East-Northeast",
            "East",
            "East-Southeast",
            "Southeast",
            "South-Southeast",
            "South",
            "South-Southwest",
            "Southwest",
            "West-Southwest",
            "West",
            "West-Northwest",
            "Northwest",
            "North-Northwest",
            "North",
        ]
    else:
        directions = [
            "N",
            "NNE",
            "NE",
            "ENE",
            "E",
            "ESE",
            "SE",
            "SSE",
            "S",
            "SSW",
            "SW",
            "WSW",
            "W",
            "WNW",
            "NW",
            "NNW",
            "N",
        ]

    cardinal = directions[round(bearing / 22.5)]
    return cardinal



import logging
import time

import requests
from oslo_config import cfg

import aprsd
from aprsd import threads as aprsd_threads

CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class APRSRegistryThread(aprsd_threads.APRSDThread):
    """This sends service information to the configured APRS Registry."""

    _loop_cnt: int = 1

    def __init__(self):
        super().__init__("APRSRegistryThread")
        self._loop_cnt = 1
        if not CONF.aprs_registry.enabled:
            LOG.error(
                "APRS Registry is not enabled.  ",
            )
            LOG.error(
                "APRS Registry thread is STOPPING.",
            )
            self.stop()
        LOG.info(
            "APRS Registry thread is running and will send "
            f"info every {CONF.aprs_registry.frequency_seconds} seconds "
            f"to {CONF.aprs_registry.registry_url}.",
        )

    def loop(self):
        # Only call the registry every N seconds
        if self._loop_cnt % CONF.aprs_registry.frequency_seconds == 0:
            info = {
                "callsign": CONF.callsign,
                "description": CONF.aprs_registry.description,
                "service_website": CONF.aprs_registry.service_website,
                "software": f"APRSD version {aprsd.__version__} "
                "https://github.com/craigerl/aprsd",
            }
            try:
                requests.post(
                    f"{CONF.aprs_registry.registry_url}",
                    json=info,
                )
            except Exception as e:
                LOG.error(f"Failed to send registry info: {e}")

        time.sleep(1)
        self._loop_cnt += 1
        return True



import inspect
import logging

import click
from rich.console import Console
from rich.table import Table
from rich.text import Text

from aprsd import cli_helper
from aprsd import plugin as aprsd_plugin
from aprsd.main import cli
from aprsd.plugins import fortune, notify, ping, time, version, weather
from aprsd.utils import package as aprsd_package

LOG = logging.getLogger('APRSD')


def show_built_in_plugins(console):
    modules = [fortune, notify, ping, time, version, weather]
    plugins = []

    for module in modules:
        entries = inspect.getmembers(module, inspect.isclass)
        for entry in entries:
            cls = entry[1]
            if issubclass(cls, aprsd_plugin.APRSDPluginBase):
                info = {
                    'name': cls.__qualname__,
                    'path': f'{cls.__module__}.{cls.__qualname__}',
                    'version': cls.version,
                    'docstring': cls.__doc__,
                    'short_desc': cls.short_description,
                }

                if issubclass(cls, aprsd_plugin.APRSDRegexCommandPluginBase):
                    info['command_regex'] = cls.command_regex
                    info['type'] = 'RegexCommand'

                if issubclass(cls, aprsd_plugin.APRSDWatchListPluginBase):
                    info['type'] = 'WatchList'

                plugins.append(info)

    plugins = sorted(plugins, key=lambda i: i['name'])

    table = Table(
        title='[not italic]:snake:[/] [bold][magenta]APRSD Built-in Plugins [not italic]:snake:[/]',
    )
    table.add_column('Plugin Name', style='cyan', no_wrap=True)
    table.add_column('Info', style='bold yellow')
    table.add_column('Type', style='bold green')
    table.add_column('Plugin Path', style='bold blue')
    for entry in plugins:
        table.add_row(entry['name'], entry['short_desc'], entry['type'], entry['path'])

    console.print(table)


def show_pypi_plugins(installed_plugins, console):
    packages = aprsd_package.get_pypi_packages()

    title = Text.assemble(
        ('Pypi.org APRSD Installable Plugin Packages\n\n', 'bold magenta'),
        ('Install any of the following plugins with\n', 'bold yellow'),
        ("'pip install ", 'bold white'),
        ("<Plugin Package Name>'", 'cyan'),
    )

    table = Table(title=title)
    table.add_column('Plugin Package Name', style='cyan', no_wrap=True)
    table.add_column('Description', style='yellow')
    table.add_column('Version', style='yellow', justify='center')
    table.add_column('Released', style='bold green', justify='center')
    table.add_column('Installed?', style='red', justify='center')
    emoji = ':open_file_folder:'
    for package in packages:
        link = package['info']['package_url']
        version = package['info']['version']
        package_name = package['info']['name']
        description = package['info']['summary']
        created = package['releases'][version][0]['upload_time']

        if 'aprsd-' not in package_name or '-plugin' not in package_name:
            continue

        under = package_name.replace('-', '_')
        installed = 'Yes' if under in installed_plugins else 'No'
        table.add_row(
            f'[link={link}]{emoji}[/link] {package_name}',
            description,
            version,
            created,
            installed,
        )

    console.print('\n')
    console.print(table)


def show_pypi_extensions(installed_extensions, console):
    packages = aprsd_package.get_pypi_packages()

    title = Text.assemble(
        ('Pypi.org APRSD Installable Extension Packages\n\n', 'bold magenta'),
        ('Install any of the following extensions by running\n', 'bold yellow'),
        ("'pip install ", 'bold white'),
        ("<Plugin Package Name>'", 'cyan'),
    )
    table = Table(title=title)
    table.add_column('Extension Package Name', style='cyan', no_wrap=True)
    table.add_column('Description', style='yellow')
    table.add_column('Version', style='yellow', justify='center')
    table.add_column('Released', style='bold green', justify='center')
    table.add_column('Installed?', style='red', justify='center')
    emoji = ':open_file_folder:'

    for package in packages:
        link = package['info']['package_url']
        version = package['info']['version']
        package_name = package['info']['name']
        description = package['info']['summary']
        created = package['releases'][version][0]['upload_time']
        if 'aprsd-' not in package_name or '-extension' not in package_name:
            continue

        under = package_name.replace('-', '_')
        installed = 'Yes' if under in installed_extensions else 'No'
        table.add_row(
            f'[link={link}]{emoji}[/link] {package_name}',
            description,
            version,
            created,
            installed,
        )

    console.print('\n')
    console.print(table)


def show_installed_plugins(installed_plugins, console):
    if not installed_plugins:
        return

    table = Table(
        title='[not italic]:snake:[/] [bold][magenta]APRSD Installed 3rd party Plugins [not italic]:snake:[/]',
    )
    table.add_column('Package Name', style=' bold white', no_wrap=True)
    table.add_column('Plugin Name', style='cyan', no_wrap=True)
    table.add_column('Version', style='yellow', justify='center')
    table.add_column('Type', style='bold green')
    table.add_column('Plugin Path', style='bold blue')
    for name in installed_plugins:
        for plugin in installed_plugins[name]:
            table.add_row(
                name.replace('_', '-'),
                plugin['name'],
                plugin['version'],
                aprsd_package.plugin_type(plugin['obj']),
                plugin['path'],
            )

    console.print('\n')
    console.print(table)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options_no_config
def list_plugins(ctx):
    """List the built in plugins available to APRSD."""
    console = Console()

    with console.status('Show Built-in Plugins') as status:
        show_built_in_plugins(console)

        status.update('Fetching pypi.org plugins')
        installed_plugins = aprsd_package.get_installed_plugins()
        show_pypi_plugins(installed_plugins, console)

        status.update('Looking for installed APRSD plugins')
        show_installed_plugins(installed_plugins, console)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options_no_config
def list_extensions(ctx):
    """List the built in plugins available to APRSD."""
    console = Console()

    with console.status('Show APRSD Extensions') as status:
        status.update('Fetching pypi.org APRSD Extensions')

        status.update('Looking for installed APRSD Extensions')
        installed_extensions = aprsd_package.get_installed_extensions()
        show_pypi_extensions(installed_extensions, console)



import datetime
import logging
import select
import socket
import threading

import aprslib
import wrapt
from aprslib import is_py3
from aprslib.exceptions import (
    ConnectionDrop,
    ConnectionError,
    GenericError,
    LoginError,
    ParseError,
    UnknownFormat,
)

import aprsd
from aprsd.packets import core

LOG = logging.getLogger('APRSD')


class APRSLibClient(aprslib.IS):
    """Extend the aprslib class so we can exit properly.

    This is a modified version of the aprslib.IS class that adds a stop method to
    allow the client to exit cleanly.

    The aprsis driver uses this class to connect to the APRS-IS server.
    """

    # flag to tell us to stop
    thread_stop = False

    # date for last time we heard from the server
    aprsd_keepalive = datetime.datetime.now()

    # Which server we are connected to?
    server_string = 'None'

    # timeout in seconds
    select_timeout = 1
    lock = threading.Lock()

    def stop(self):
        self.thread_stop = True
        LOG.warning('Shutdown Aprsdis client.')

    def close(self):
        LOG.warning('Closing Aprsdis client.')
        super().close()

    @wrapt.synchronized(lock)
    def send(self, packet: core.Packet):
        """Send an APRS Message object."""
        self.sendall(packet.raw)

    def is_alive(self):
        """If the connection is alive or not."""
        return self._connected

    def _connect(self):
        """
        Attemps connection to the server
        """

        self.logger.info(
            'Attempting connection to %s:%s', self.server[0], self.server[1]
        )

        try:
            self._open_socket()

            peer = self.sock.getpeername()

            self.logger.info('Connected to %s', str(peer))

            # 5 second timeout to receive server banner
            self.sock.setblocking(1)
            self.sock.settimeout(5)

            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # MACOS doesn't have TCP_KEEPIDLE
            if hasattr(socket, 'TCP_KEEPIDLE'):
                self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 1)
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 3)
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)

            banner = self.sock.recv(512)
            if is_py3:
                banner = banner.decode('latin-1')

            if banner[0] == '#':
                self.logger.debug('Banner: %s', banner.rstrip())
            else:
                raise ConnectionError('invalid banner from server')

        except ConnectionError as e:
            self.logger.error(str(e))
            self.close()
            raise
        except (socket.error, socket.timeout) as e:
            self.close()

            self.logger.error('Socket error: %s' % str(e))
            if str(e) == 'timed out':
                raise ConnectionError('no banner from server') from e
            else:
                raise ConnectionError(e) from e

        self._connected = True

    def _socket_readlines(self, blocking=False):
        """
        Generator for complete lines, received from the server
        """
        try:
            self.sock.setblocking(0)
        except OSError as e:
            self.logger.error(f'socket error when setblocking(0): {str(e)}')
            raise aprslib.ConnectionDrop('connection dropped') from e

        while not self.thread_stop:
            short_buf = b''
            newline = b'\r\n'

            # set a select timeout, so we get a chance to exit
            # when user hits CTRL-C
            readable, writable, exceptional = select.select(
                [self.sock],
                [],
                [],
                self.select_timeout,
            )
            if not readable:
                if not blocking:
                    break
                else:
                    continue

            try:
                short_buf = self.sock.recv(4096)

                # sock.recv returns empty if the connection drops
                if not short_buf:
                    if not blocking:
                        # We could just not be blocking, so empty is expected
                        continue
                    else:
                        self.logger.error('socket.recv(): returned empty')
                        raise aprslib.ConnectionDrop('connection dropped')
            except OSError as e:
                # self.logger.error("socket error on recv(): %s" % str(e))
                if 'Resource temporarily unavailable' in str(e):
                    if not blocking:
                        if len(self.buf) == 0:
                            break

            self.buf += short_buf

            while newline in self.buf:
                line, self.buf = self.buf.split(newline, 1)

                yield line

    def _send_login(self):
        """
        Sends login string to server
        """
        login_str = 'user {0} pass {1} vers Python-APRSD {3}{2}\r\n'
        login_str = login_str.format(
            self.callsign,
            self.passwd,
            (' filter ' + self.filter) if self.filter != '' else '',
            aprsd.__version__,
        )

        self.logger.debug('Sending login information')

        try:
            self._sendall(login_str)
            self.sock.settimeout(5)
            test = self.sock.recv(len(login_str) + 100)
            if is_py3:
                test = test.decode('latin-1')
            test = test.rstrip()

            self.logger.debug("Server: '%s'", test)

            if not test:
                raise LoginError(f"Server Response Empty: '{test}'")

            _, _, callsign, status, e = test.split(' ', 4)
            s = e.split(',')
            if len(s):
                server_string = s[0].replace('server ', '')
            else:
                server_string = e.replace('server ', '')

            if callsign == '':
                raise LoginError('Server responded with empty callsign???')
            if callsign != self.callsign:
                raise LoginError(f'Server: {test}')
            if status != 'verified,' and self.passwd != '-1':
                raise LoginError('Password is incorrect')

            if self.passwd == '-1':
                self.logger.info('Login successful (receive only)')
            else:
                self.logger.info('Login successful')

            self.logger.info(f'Connected to {server_string}')
            self.server_string = server_string

        except LoginError as e:
            self.logger.error(str(e))
            self.close()
            raise
        except Exception as e:
            self.close()
            self.logger.error(f"Failed to login '{e}'")
            self.logger.exception(e)
            raise LoginError('Failed to login') from e

    def consumer(self, callback, blocking=True, immortal=False, raw=False):
        """
        When a position sentence is received, it will be passed to the callback function

        blocking: if true (default), runs forever, otherwise will return after one sentence
                  You can still exit the loop, by raising StopIteration in the callback function

        immortal: When true, consumer will try to reconnect and stop propagation of Parse exceptions
                  if false (default), consumer will return

        raw: when true, raw packet is passed to callback, otherwise the result from aprs.parse()
        """

        if not self._connected:
            raise ConnectionError('not connected to a server')

        line = b''

        while not self.thread_stop:
            try:
                for line in self._socket_readlines(blocking):
                    if line[0:1] != b'#':
                        self.aprsd_keepalive = datetime.datetime.now()
                        if raw:
                            callback(line)
                        else:
                            callback(self._parse(line))
                    else:
                        self.logger.debug('Server: %s', line.decode('utf8'))
                        self.aprsd_keepalive = datetime.datetime.now()
            except ParseError as exp:
                self.logger.log(
                    11,
                    "%s    Packet: '%s'",
                    exp,
                    exp.packet,
                )
            except UnknownFormat as exp:
                self.logger.log(
                    9,
                    "%s    Packet: '%s'",
                    exp,
                    exp.packet,
                )
            except LoginError as exp:
                self.logger.error('%s: %s', exp.__class__.__name__, exp)
            except (KeyboardInterrupt, SystemExit):
                raise
            except (ConnectionDrop, ConnectionError):
                self.close()

                if not immortal:
                    raise
                else:
                    self.connect(blocking=blocking)
                    continue
            except GenericError:
                pass
            except StopIteration:
                break
            except IOError:
                if not self.thread_stop:
                    self.logger.error('IOError')
                break
            except Exception:
                self.logger.error('APRS Packet: %s', line)
                raise

            if not blocking:
                break



import logging
from typing import Callable, Protocol, runtime_checkable

from aprsd.utils import singleton

LOG = logging.getLogger('APRSD')


@runtime_checkable
class KeepAliveProducer(Protocol):
    """The KeepAliveProducer protocol is used to define the interface for running Keepalive checks."""

    def keepalive_check(self) -> dict:
        """Check for keepalive."""
        ...

    def keepalive_log(self):
        """Log any keepalive information."""
        ...


@singleton
class KeepAliveCollector:
    """The Collector class is used to collect stats from multiple StatsProducer instances."""

    def __init__(self):
        self.producers: list[Callable] = []

    def check(self) -> None:
        """Do any keepalive checks."""
        for name in self.producers:
            cls = name()
            try:
                cls.keepalive_check()
            except Exception as e:
                LOG.error(f'Error in producer {name} (check): {e}')
                raise e

    def log(self) -> None:
        """Log any relevant information during a KeepAlive check"""
        for name in self.producers:
            cls = name()
            try:
                cls.keepalive_log()
            except Exception as e:
                LOG.error(f'Error in producer {name} (check): {e}')
                raise e

    def register(self, producer_name: Callable):
        if not isinstance(producer_name, KeepAliveProducer):
            raise TypeError(f'Producer {producer_name} is not a KeepAliveProducer')
        self.producers.append(producer_name)

    def unregister(self, producer_name: Callable):
        if not isinstance(producer_name, KeepAliveProducer):
            raise TypeError(f'Producer {producer_name} is not a KeepAliveProducer')
        self.producers.remove(producer_name)



import abc
import datetime
import logging
import threading
import time
from typing import List

import wrapt

LOG = logging.getLogger("APRSD")


class APRSDThread(threading.Thread, metaclass=abc.ABCMeta):
    """Base class for all threads in APRSD."""

    loop_count = 1
    _pause = False
    thread_stop = False

    def __init__(self, name):
        super().__init__(name=name)
        self.thread_stop = False
        APRSDThreadList().add(self)
        self._last_loop = datetime.datetime.now()

    def _should_quit(self):
        """see if we have a quit message from the global queue."""
        if self.thread_stop:
            return True

    def pause(self):
        """Logically pause the processing of the main loop."""
        LOG.debug(f"Pausing thread '{self.name}' loop_count {self.loop_count}")
        self._pause = True

    def unpause(self):
        """Logically resume processing of the main loop."""
        LOG.debug(f"Resuming thread '{self.name}' loop_count {self.loop_count}")
        self._pause = False

    def stop(self):
        LOG.debug(f"Stopping thread '{self.name}'")
        self.thread_stop = True

    @abc.abstractmethod
    def loop(self):
        pass

    def _cleanup(self):
        """Add code to subclass to do any cleanup"""

    def __str__(self):
        out = (
            f"Thread <{self.__class__.__name__}({self.name}) Alive? {self.is_alive()}>"
        )
        return out

    def loop_age(self):
        """How old is the last loop call?"""
        return datetime.datetime.now() - self._last_loop

    def run(self):
        LOG.debug("Starting")
        while not self._should_quit():
            if self._pause:
                time.sleep(1)
            else:
                self.loop_count += 1
                can_loop = self.loop()
                self._last_loop = datetime.datetime.now()
                if not can_loop:
                    self.stop()
        self._cleanup()
        APRSDThreadList().remove(self)
        LOG.debug("Exiting")


class APRSDThreadList:
    """Singleton class that keeps track of application wide threads."""

    _instance = None

    threads_list: List[APRSDThread] = []
    lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls.threads_list = []
        return cls._instance

    def __contains__(self, name):
        """See if we have a thread in our list"""
        for t in self.threads_list:
            if t.name == name:
                return True
        return False

    def stats(self, serializable=False) -> dict:
        stats = {}
        for th in self.threads_list:
            age = th.loop_age()
            if serializable:
                age = str(age)
            stats[th.name] = {
                "name": th.name,
                "class": th.__class__.__name__,
                "alive": th.is_alive(),
                "age": th.loop_age(),
                "loop_count": th.loop_count,
            }
        return stats

    @wrapt.synchronized(lock)
    def add(self, thread_obj):
        self.threads_list.append(thread_obj)

    @wrapt.synchronized(lock)
    def remove(self, thread_obj):
        self.threads_list.remove(thread_obj)

    @wrapt.synchronized(lock)
    def stop_all(self):
        """Iterate over all threads and call stop on them."""
        for th in self.threads_list:
            LOG.info(f"Stopping Thread {th.name}")
            if hasattr(th, "packet"):
                LOG.info(f"{th.name} packet {th.packet}")
            th.stop()

    @wrapt.synchronized
    def pause_all(self):
        """Iterate over all threads and pause them."""
        for th in self.threads_list:
            LOG.info(f"Pausing Thread {th.name}")
            if hasattr(th, "packet"):
                LOG.info(f"{th.name} packet {th.packet}")
            th.pause()

    @wrapt.synchronized
    def unpause_all(self):
        """Iterate over all threads and resume them."""
        for th in self.threads_list:
            LOG.info(f"Resuming Thread {th.name}")
            if hasattr(th, "packet"):
                LOG.info(f"{th.name} packet {th.packet}")
            th.unpause()

    @wrapt.synchronized(lock)
    def info(self):
        """Go through all the threads and collect info about each."""
        info = {}
        for thread in self.threads_list:
            alive = thread.is_alive()
            age = thread.loop_age()
            key = thread.__class__.__name__
            info[key] = {
                "alive": True if alive else False,
                "age": age,
                "name": thread.name,
            }
        return info

    @wrapt.synchronized(lock)
    def __len__(self):
        return len(self.threads_list)



import datetime
import logging
import time
from typing import Callable

from aprslib.exceptions import LoginError
from loguru import logger
from oslo_config import cfg

from aprsd import client, exception
from aprsd.client.drivers.lib.aprslib import APRSLibClient
from aprsd.packets import core
from aprsd.utils import singleton

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')
LOGU = logger


# class APRSISDriver(metaclass=trace.TraceWrapperMetaclass):
@singleton
class APRSISDriver:
    """This is the APRS-IS driver for the APRSD client.

    This driver uses our modified aprslib.IS class to connect to the APRS-IS server.

    """

    _client = None
    _checks = False
    connected = False

    def __init__(self):
        max_timeout = {'hours': 0.0, 'minutes': 2, 'seconds': 0}
        self.max_delta = datetime.timedelta(**max_timeout)
        self.login_status = {
            'success': False,
            'message': None,
        }

    @staticmethod
    def is_enabled():
        # Defaults to True if the enabled flag is non existent
        try:
            return CONF.aprs_network.enabled
        except KeyError:
            return False

    @staticmethod
    def is_configured():
        if APRSISDriver.is_enabled():
            # Ensure that the config vars are correctly set
            if not CONF.aprs_network.login:
                LOG.error('Config aprs_network.login not set.')
                raise exception.MissingConfigOptionException(
                    'aprs_network.login is not set.',
                )
            if not CONF.aprs_network.password:
                LOG.error('Config aprs_network.password not set.')
                raise exception.MissingConfigOptionException(
                    'aprs_network.password is not set.',
                )
            if not CONF.aprs_network.host:
                LOG.error('Config aprs_network.host not set.')
                raise exception.MissingConfigOptionException(
                    'aprs_network.host is not set.',
                )

            return True
        return True

    @property
    def is_alive(self):
        if not self._client:
            LOG.warning(f'APRS_CLIENT {self._client} alive? NO!!!')
            return False
        return self._client.is_alive() and not self._is_stale_connection()

    def close(self):
        if self._client:
            self._client.stop()
            self._client.close()
        self.connected = False

    def send(self, packet: core.Packet) -> bool:
        return self._client.send(packet)

    def setup_connection(self):
        if self.connected:
            return
        user = CONF.aprs_network.login
        password = CONF.aprs_network.password
        host = CONF.aprs_network.host
        port = CONF.aprs_network.port
        backoff = 1
        retries = 3
        retry_count = 0
        while not self.connected:
            retry_count += 1
            if retry_count >= retries:
                break
            try:
                LOG.info(
                    f'Creating aprslib client({host}:{port}) and logging in {user}. try #{retry_count}'
                )
                self._client = APRSLibClient(
                    user, passwd=password, host=host, port=port
                )
                # Force the log to be the same
                self._client.logger = LOG
                self._client.connect()
                self.connected = self.login_status['success'] = True
                self.login_status['message'] = self._client.server_string
                backoff = 1
            except LoginError as e:
                LOG.error(f"Failed to login to APRS-IS Server '{e}'")
                self.connected = self.login_status['success'] = False
                self.login_status['message'] = (
                    e.message if hasattr(e, 'message') else str(e)
                )
                LOG.error(self.login_status['message'])
                time.sleep(backoff)
            except Exception as e:
                LOG.error(f"Unable to connect to APRS-IS server. '{e}' ")
                self.connected = self.login_status['success'] = False
                self.login_status['message'] = getattr(e, 'message', str(e))
                time.sleep(backoff)
                # Don't allow the backoff to go to inifinity.
                if backoff > 5:
                    backoff = 5
                else:
                    backoff += 1
                continue

    def set_filter(self, filter):
        self._client.set_filter(filter)

    def login_success(self) -> bool:
        return self.login_status.get('success', False)

    def login_failure(self) -> str:
        return self.login_status.get('message', None)

    @property
    def filter(self):
        return self._client.filter

    @property
    def server_string(self):
        return self._client.server_string

    @property
    def keepalive(self):
        return self._client.aprsd_keepalive

    def _is_stale_connection(self):
        delta = datetime.datetime.now() - self._client.aprsd_keepalive
        if delta > self.max_delta:
            LOG.warning(f'Connection is stale, last heard {delta} ago.')
            return True
        return False

    @staticmethod
    def transport():
        return client.TRANSPORT_APRSIS

    def decode_packet(self, *args, **kwargs):
        """APRS lib already decodes this."""
        return core.factory(args[0])

    def consumer(self, callback: Callable, raw: bool = False):
        if self._client and self.connected:
            try:
                self._client.consumer(
                    callback,
                    blocking=False,
                    immortal=False,
                    raw=raw,
                )
            except Exception as e:
                LOG.error(e)
                LOG.info(e.__cause__)
                raise e
        else:
            self.connected = False

    def stats(self, serializable: bool = False) -> dict:
        stats = {}
        if self.is_configured():
            if self._client:
                keepalive = self._client.aprsd_keepalive
                server_string = self._client.server_string
                if serializable:
                    keepalive = keepalive.isoformat()
                filter = self.filter
            else:
                keepalive = 'None'
                server_string = 'None'
                filter = 'None'
            stats = {
                'connected': self.is_alive,
                'filter': filter,
                'login_status': self.login_status,
                'connection_keepalive': keepalive,
                'server_string': server_string,
                'transport': self.transport(),
            }

        return stats



import logging
from typing import Callable, Protocol, runtime_checkable, Union, Dict

from aprsd.packets import core
from aprsd.utils import singleton

LOG = logging.getLogger("APRSD")


@runtime_checkable
class PacketFilterProtocol(Protocol):
    """Protocol API for a packet filter class.
    """
    def filter(self, packet: type[core.Packet]) -> Union[type[core.Packet], None]:
        """When we get a packet from the network.

        Return a Packet object if the filter passes. Return None if the
        Packet is filtered out.
        """
        ...


@singleton
class PacketFilter:

    def __init__(self):
        self.filters: Dict[str, Callable] = {}

    def register(self, packet_filter: Callable) -> None:
        if not isinstance(packet_filter, PacketFilterProtocol):
            raise TypeError(f"class {packet_filter} is not a PacketFilterProtocol object")

        if packet_filter not in self.filters:
            self.filters[packet_filter] = packet_filter()

    def unregister(self, packet_filter: Callable) -> None:
        if not isinstance(packet_filter, PacketFilterProtocol):
            raise TypeError(f"class {packet_filter} is not a PacketFilterProtocol object")
        if packet_filter in self.filters:
            del self.filters[packet_filter]

    def filter(self, packet: type[core.Packet]) -> Union[type[core.Packet], None]:
        """Run through each of the filters.

        This will step through each registered filter class
        and call filter on it.

        If the filter object returns None, we are done filtering.
        If the filter object returns the packet, we continue filtering.
        """
        for packet_filter in self.filters:
            try:
                if not self.filters[packet_filter].filter(packet):
                    LOG.debug(f"{self.filters[packet_filter].__class__.__name__} dropped {packet.__class__.__name__}:{packet.human_info}")
                    return None
            except Exception as ex:
                LOG.error(f"{packet_filter.__clas__.__name__} failed filtering packet {packet.__class__.__name__} : {ex}")
        return packet



from unittest import mock

from oslo_config import cfg

from aprsd import (  # noqa: F401
    client,
    conf,
    packets,
)
from aprsd.client.drivers.registry import DriverRegistry
from aprsd.plugins import notify as notify_plugin

from .. import fake, test_plugin
from ..mock_client_driver import MockClientDriver

CONF = cfg.CONF
DEFAULT_WATCHLIST_CALLSIGNS = fake.FAKE_FROM_CALLSIGN


class TestWatchListPlugin(test_plugin.TestPlugin):
    def setUp(self):
        super().setUp()
        self.fromcall = fake.FAKE_FROM_CALLSIGN
        self.ack = 1

        # Mock APRSISDriver
        self.aprsis_patcher = mock.patch('aprsd.client.drivers.aprsis.APRSISDriver')
        self.mock_aprsis = self.aprsis_patcher.start()
        self.mock_aprsis.is_enabled.return_value = False
        self.mock_aprsis.is_configured.return_value = False

        # Register the mock driver
        DriverRegistry().register(MockClientDriver)

    def tearDown(self):
        super().tearDown()
        if hasattr(self, 'aprsis_patcher'):
            self.aprsis_patcher.stop()

    def config_and_init(
        self,
        watchlist_enabled=True,
        watchlist_alert_callsign=None,
        watchlist_alert_time_seconds=None,
        watchlist_packet_keep_count=None,
        watchlist_callsigns=DEFAULT_WATCHLIST_CALLSIGNS,
    ):
        CONF.callsign = self.fromcall
        CONF.aprs_network.login = self.fromcall
        CONF.aprs_fi.apiKey = 'something'
        # Add mock password
        CONF.aprs_network.password = '12345'

        # Set the watchlist specific config options
        CONF.watch_list.enabled = watchlist_enabled

        if not watchlist_alert_callsign:
            watchlist_alert_callsign = fake.FAKE_TO_CALLSIGN
        CONF.watch_list.alert_callsign = watchlist_alert_callsign

        if not watchlist_alert_time_seconds:
            watchlist_alert_time_seconds = CONF.watch_list.alert_time_seconds
        CONF.watch_list.alert_time_seconds = watchlist_alert_time_seconds

        if not watchlist_packet_keep_count:
            watchlist_packet_keep_count = CONF.watch_list.packet_keep_count
            CONF.watch_list.packet_keep_count = watchlist_packet_keep_count

        CONF.watch_list.callsigns = watchlist_callsigns


class TestAPRSDWatchListPluginBase(TestWatchListPlugin):
    def test_watchlist_not_enabled(self):
        self.config_and_init(watchlist_enabled=False)
        plugin = fake.FakeWatchListPlugin()

        packet = fake.fake_packet(
            message='version',
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    def test_watchlist_not_in_watchlist(self):
        self.config_and_init()
        plugin = fake.FakeWatchListPlugin()

        packet = fake.fake_packet(
            fromcall='FAKE',
            message='version',
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)


class TestNotifySeenPlugin(TestWatchListPlugin):
    def test_disabled(self):
        self.config_and_init(watchlist_enabled=False)
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            message='version',
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    def test_callsign_not_in_watchlist(self):
        self.config_and_init(watchlist_enabled=False)
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            message='version',
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch('aprsd.packets.WatchList.is_old')
    def test_callsign_in_watchlist_not_old(self, mock_is_old):
        mock_is_old.return_value = False
        self.config_and_init(
            watchlist_enabled=True,
            watchlist_callsigns=['WB4BOR'],
        )
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            fromcall='WB4BOR',
            message='ping',
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch('aprsd.packets.WatchList.is_old')
    def test_callsign_in_watchlist_old_same_alert_callsign(self, mock_is_old):
        mock_is_old.return_value = True
        self.config_and_init(
            watchlist_enabled=True,
            watchlist_alert_callsign='WB4BOR',
            watchlist_callsigns=['WB4BOR'],
        )
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            fromcall='WB4BOR',
            message='ping',
            msg_number=1,
        )
        actual = plugin.filter(packet)
        expected = packets.NULL_MESSAGE
        self.assertEqual(expected, actual)

    @mock.patch('aprsd.packets.WatchList.is_old')
    def test_callsign_in_watchlist_old_send_alert(self, mock_is_old):
        mock_is_old.return_value = True
        notify_callsign = fake.FAKE_TO_CALLSIGN
        fromcall = 'WB4BOR'
        self.config_and_init(
            watchlist_enabled=True,
            watchlist_alert_callsign=notify_callsign,
            watchlist_callsigns=['WB4BOR'],
        )
        plugin = notify_plugin.NotifySeenPlugin()

        packet = fake.fake_packet(
            fromcall=fromcall,
            message='ping',
            msg_number=1,
        )
        packet_type = packet.__class__.__name__
        actual = plugin.filter(packet)
        msg = f"{fromcall} was just seen by type:'{packet_type}'"

        self.assertIsInstance(actual, packets.MessagePacket)
        self.assertEqual(fake.FAKE_FROM_CALLSIGN, actual.from_call)
        self.assertEqual(notify_callsign, actual.to_call)
        self.assertEqual(msg, actual.message_text)



import logging
import re

from oslo_config import cfg
import pytz
from tzlocal import get_localzone

from aprsd import packets, plugin, plugin_utils
from aprsd.utils import fuzzy, trace


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


class TimePlugin(plugin.APRSDRegexCommandPluginBase):
    """Time command."""

    # Look for t or t<space> or T<space> or time
    command_regex = r"^([t]|[t]\s|time)"
    command_name = "time"
    short_description = "What is the current local time."

    def _get_local_tz(self):
        lz = get_localzone()
        return pytz.timezone(str(lz))

    def _get_utcnow(self):
        return pytz.datetime.datetime.utcnow()

    def build_date_str(self, localzone):
        utcnow = self._get_utcnow()
        gmt_t = pytz.utc.localize(utcnow)
        local_t = gmt_t.astimezone(localzone)

        local_short_str = local_t.strftime("%H:%M %Z")
        local_hour = local_t.strftime("%H")
        local_min = local_t.strftime("%M")
        cur_time = fuzzy(int(local_hour), int(local_min), 1)

        reply = "{} ({})".format(
            cur_time,
            local_short_str,
        )

        return reply

    @trace.trace
    def process(self, packet: packets.Packet):
        LOG.info("TIME COMMAND")
        # So we can mock this in unit tests
        localzone = self._get_local_tz()
        return self.build_date_str(localzone)


class TimeOWMPlugin(TimePlugin, plugin.APRSFIKEYMixin):
    """OpenWeatherMap based timezone fetching."""

    command_regex = r"^([t]|[t]\s|time)"
    command_name = "time"
    short_description = "Current time of GPS beacon's timezone. Uses OpenWeatherMap"

    def setup(self):
        self.ensure_aprs_fi_key()

    @trace.trace
    def process(self, packet: packets.MessagePacket):
        fromcall = packet.from_call
        message = packet.message_text
        # ack = packet.get("msgNo", "0")

        # optional second argument is a callsign to search
        a = re.search(r"^.*\s+(.*)", message)
        if a is not None:
            searchcall = a.group(1)
            searchcall = searchcall.upper()
        else:
            # if no second argument, search for calling station
            searchcall = fromcall

        api_key = CONF.aprs_fi.apiKey
        try:
            aprs_data = plugin_utils.get_aprs_fi(api_key, searchcall)
        except Exception as ex:
            LOG.error(f"Failed to fetch aprs.fi data {ex}")
            return "Failed to fetch location"

        LOG.debug(f"LocationPlugin: aprs_data = {aprs_data}")
        if not len(aprs_data["entries"]):
            LOG.error("Didn't get any entries from aprs.fi")
            return "Failed to fetch aprs.fi location"

        lat = aprs_data["entries"][0]["lat"]
        lon = aprs_data["entries"][0]["lng"]

        try:
            self.config.exists(
                ["services", "openweathermap", "apiKey"],
            )
        except Exception as ex:
            LOG.error(f"Failed to find config openweathermap:apiKey {ex}")
            return "No openweathermap apiKey found"

        api_key = self.config["services"]["openweathermap"]["apiKey"]
        try:
            results = plugin_utils.fetch_openweathermap(api_key, lat, lon)
        except Exception as ex:
            LOG.error(f"Couldn't fetch openweathermap api '{ex}'")
            # default to UTC
            localzone = pytz.timezone("UTC")
        else:
            tzone = results["timezone"]
            localzone = pytz.timezone(tzone)

        return self.build_date_str(localzone)



import logging
import shutil
import subprocess

from aprsd import packets, plugin
from aprsd.utils import trace

LOG = logging.getLogger('APRSD')

FORTUNE_PATHS = [
    '/usr/games/fortune',
    '/usr/local/bin/fortune',
    '/usr/bin/fortune',
]


class FortunePlugin(plugin.APRSDRegexCommandPluginBase):
    """Fortune."""

    command_regex = r'^([f]|[f]\s|fortune)'
    command_name = 'fortune'
    short_description = 'Give me a fortune'

    fortune_path = None

    def setup(self):
        for path in FORTUNE_PATHS:
            self.fortune_path = shutil.which(path)
            LOG.info(f'Fortune path {self.fortune_path}')
            if self.fortune_path:
                break
        if not self.fortune_path:
            self.enabled = False
        else:
            self.enabled = True

    @trace.trace
    def process(self, packet: packets.MessagePacket):
        LOG.info('FortunePlugin')

        # fromcall = packet.get("from")
        # message = packet.get("message_text", None)
        # ack = packet.get("msgNo", "0")

        reply = None

        try:
            cmnd = [self.fortune_path, '-s', '-n 60']
            command = ' '.join(cmnd)
            output = subprocess.check_output(
                command,
                shell=True,
                timeout=3,
                text=True,
            )
            output = (
                output.replace('\r', '')
                .replace('\n', '')
                .replace('  ', '')
                .replace('\t', ' ')
            )
        except subprocess.CalledProcessError as ex:
            reply = f"Fortune command failed '{ex.output}'"
        else:
            reply = output

        return reply



import datetime
import logging
import time
import tracemalloc

from loguru import logger
from oslo_config import cfg

from aprsd import packets, utils
from aprsd.log import log as aprsd_log
from aprsd.stats import collector
from aprsd.threads import APRSDThread, APRSDThreadList
from aprsd.utils import keepalive_collector

CONF = cfg.CONF
LOG = logging.getLogger("APRSD")
LOGU = logger


class KeepAliveThread(APRSDThread):
    cntr = 0
    checker_time = datetime.datetime.now()

    def __init__(self):
        tracemalloc.start()
        super().__init__("KeepAlive")
        max_timeout = {"hours": 0.0, "minutes": 2, "seconds": 0}
        self.max_delta = datetime.timedelta(**max_timeout)

    def loop(self):
        if self.loop_count % 60 == 0:
            stats_json = collector.Collector().collect()
            pl = packets.PacketList()
            thread_list = APRSDThreadList()
            now = datetime.datetime.now()

            if (
                "APRSClientStats" in stats_json
                and stats_json["APRSClientStats"].get("transport") == "aprsis"
            ):
                if stats_json["APRSClientStats"].get("server_keepalive"):
                    last_msg_time = utils.strfdelta(
                        now - stats_json["APRSClientStats"]["server_keepalive"]
                    )
                else:
                    last_msg_time = "N/A"
            else:
                last_msg_time = "N/A"

            tracked_packets = stats_json["PacketTrack"]["total_tracked"]
            tx_msg = 0
            rx_msg = 0
            if "PacketList" in stats_json:
                msg_packets = stats_json["PacketList"].get("MessagePacket")
                if msg_packets:
                    tx_msg = msg_packets.get("tx", 0)
                    rx_msg = msg_packets.get("rx", 0)

            keepalive = (
                "{} - Uptime {} RX:{} TX:{} Tracker:{} Msgs TX:{} RX:{} "
                "Last:{} - RAM Current:{} Peak:{} Threads:{} LoggingQueue:{}"
            ).format(
                stats_json["APRSDStats"]["callsign"],
                stats_json["APRSDStats"]["uptime"],
                pl.total_rx(),
                pl.total_tx(),
                tracked_packets,
                tx_msg,
                rx_msg,
                last_msg_time,
                stats_json["APRSDStats"]["memory_current_str"],
                stats_json["APRSDStats"]["memory_peak_str"],
                len(thread_list),
                aprsd_log.logging_queue.qsize(),
            )
            LOG.info(keepalive)
            if "APRSDThreadList" in stats_json:
                thread_list = stats_json["APRSDThreadList"]
                for thread_name in thread_list:
                    thread = thread_list[thread_name]
                    alive = thread["alive"]
                    age = thread["age"]
                    key = thread["name"]
                    if not alive:
                        LOG.error(f"Thread {thread}")

                    thread_hex = f"fg {utils.hex_from_name(key)}"
                    t_name = f"<{thread_hex}>{key:<15}</{thread_hex}>"
                    thread_msg = f"{t_name} Alive? {str(alive): <5} {str(age): <20}"
                    LOGU.opt(colors=True).info(thread_msg)
                    # LOG.info(f"{key: <15} Alive? {str(alive): <5} {str(age): <20}")

            # Go through the registered keepalive collectors
            # and check them as well as call log.
            collect = keepalive_collector.KeepAliveCollector()
            collect.check()
            collect.log()

            # Check version every day
            delta = now - self.checker_time
            if delta > datetime.timedelta(hours=24):
                self.checker_time = now
                level, msg = utils._check_version()
                if level:
                    LOG.warning(msg)
            self.cntr += 1
        time.sleep(1)
        return True



# aprsd/aprsd/threads/service.py
#
# This module is used to register threads that the service command runs.
#
# The service command is used to start and stop the APRS service.
# This is a mechanism to register threads that the service or command
# needs to run, and then start stop them as needed.

from aprsd.threads import aprsd as aprsd_threads
from aprsd.utils import singleton


@singleton
class ServiceThreads:
    """Registry for threads that the service command runs.

    This enables extensions to register a thread to run during
    the service command.
    """

    def __init__(self):
        self.threads: list[aprsd_threads.APRSDThread] = []

    def register(self, thread: aprsd_threads.APRSDThread):
        if not isinstance(thread, aprsd_threads.APRSDThread):
            raise TypeError(f'Thread {thread} is not an APRSDThread')
        self.threads.append(thread)

    def unregister(self, thread: aprsd_threads.APRSDThread):
        if not isinstance(thread, aprsd_threads.APRSDThread):
            raise TypeError(f'Thread {thread} is not an APRSDThread')
        self.threads.remove(thread)

    def start(self):
        """Start all threads in the list."""
        for thread in self.threads:
            thread.start()

    def join(self):
        """Join all the threads in the list"""
        for thread in self.threads:
            thread.join()



#
# License GPLv2
#

# python included libs
import datetime
import logging
import signal
import sys
import time

import click
from loguru import logger
from oslo_config import cfg
from rich.console import Console

# local imports here
import aprsd
from aprsd import cli_helper, packets, plugin, threads, utils
from aprsd.client.client import APRSDClient
from aprsd.main import cli
from aprsd.packets import collector as packet_collector
from aprsd.packets import core, seen_list
from aprsd.packets import log as packet_log
from aprsd.packets.filter import PacketFilter
from aprsd.packets.filters import dupe_filter, packet_type
from aprsd.stats import collector
from aprsd.threads import keepalive, rx
from aprsd.threads import stats as stats_thread
from aprsd.threads.aprsd import APRSDThread

# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
LOG = logging.getLogger('APRSD')
CONF = cfg.CONF
LOGU = logger
console = Console()


def signal_handler(sig, frame):
    threads.APRSDThreadList().stop_all()
    if 'subprocess' not in str(frame):
        LOG.info(
            'Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}'.format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(5)
        # Last save to disk
        collector.Collector().collect()


class APRSDListenProcessThread(rx.APRSDFilterThread):
    def __init__(
        self,
        packet_queue,
        packet_filter=None,
        plugin_manager=None,
        enabled_plugins=None,
        log_packets=False,
    ):
        super().__init__('ListenProcThread', packet_queue)
        self.packet_filter = packet_filter
        self.plugin_manager = plugin_manager
        if self.plugin_manager:
            LOG.info(f'Plugins {self.plugin_manager.get_message_plugins()}')
        self.log_packets = log_packets

    def print_packet(self, packet):
        if self.log_packets:
            packet_log.log(packet)

    def process_packet(self, packet: type[core.Packet]):
        if self.plugin_manager:
            # Don't do anything with the reply.
            # This is the listen only command.
            self.plugin_manager.run(packet)


class ListenStatsThread(APRSDThread):
    """Log the stats from the PacketList."""

    def __init__(self):
        super().__init__('PacketStatsLog')
        self._last_total_rx = 0
        self.period = 31

    def loop(self):
        if self.loop_count % self.period == 0:
            # log the stats every 10 seconds
            stats_json = collector.Collector().collect()
            stats = stats_json['PacketList']
            total_rx = stats['rx']
            packet_count = len(stats['packets'])
            rx_delta = total_rx - self._last_total_rx
            rate = rx_delta / self.period

            # Log summary stats
            LOGU.opt(colors=True).info(
                f'<green>RX Rate: {rate:.2f} pps</green>  '
                f'<yellow>Total RX: {total_rx}</yellow> '
                f'<red>RX Last {self.period} secs: {rx_delta}</red> '
                f'<white>Packets in PacketListStats: {packet_count}</white>',
            )
            self._last_total_rx = total_rx

            # Log individual type stats
            for k, v in stats['types'].items():
                thread_hex = f'fg {utils.hex_from_name(k)}'
                LOGU.opt(colors=True).info(
                    f'<{thread_hex}>{k:<15}</{thread_hex}> '
                    f'<blue>RX: {v["rx"]}</blue> <red>TX: {v["tx"]}</red>',
                )

        time.sleep(1)
        return True


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    '--aprs-login',
    envvar='APRS_LOGIN',
    show_envvar=True,
    help='What callsign to send the message from.',
)
@click.option(
    '--aprs-password',
    envvar='APRS_PASSWORD',
    show_envvar=True,
    help='the APRS-IS password for APRS_LOGIN',
)
@click.option(
    '--packet-filter',
    type=click.Choice(
        [
            packets.AckPacket.__name__,
            packets.BeaconPacket.__name__,
            packets.GPSPacket.__name__,
            packets.MicEPacket.__name__,
            packets.MessagePacket.__name__,
            packets.ObjectPacket.__name__,
            packets.RejectPacket.__name__,
            packets.StatusPacket.__name__,
            packets.ThirdPartyPacket.__name__,
            packets.UnknownPacket.__name__,
            packets.WeatherPacket.__name__,
        ],
        case_sensitive=False,
    ),
    multiple=True,
    default=[],
    help='Filter by packet type',
)
@click.option(
    '--enable-plugin',
    multiple=True,
    help='Enable a plugin.  This is the name of the file in the plugins directory.',
)
@click.option(
    '--load-plugins',
    default=False,
    is_flag=True,
    help='Load plugins as enabled in aprsd.conf ?',
)
@click.argument(
    'filter',
    nargs=-1,
    required=True,
)
@click.option(
    '--log-packets',
    default=False,
    is_flag=True,
    help='Log incoming packets.',
)
@click.option(
    '--enable-packet-stats',
    default=False,
    is_flag=True,
    help='Enable packet stats periodic logging.',
)
@click.pass_context
@cli_helper.process_standard_options
def listen(
    ctx,
    aprs_login,
    aprs_password,
    packet_filter,
    enable_plugin,
    load_plugins,
    filter,
    log_packets,
    enable_packet_stats,
):
    """Listen to packets on the APRS-IS Network based on FILTER.

    FILTER is the APRS Filter to use.\n
     see http://www.aprs-is.net/javAPRSFilter.aspx\n
    r/lat/lon/dist - Range Filter Pass posits and objects within dist km from lat/lon.\n
    p/aa/bb/cc... - Prefix Filter Pass traffic with fromCall that start with aa or bb or cc.\n
    b/call1/call2... - Budlist Filter Pass all traffic from exact call: call1, call2, ... (* wild card allowed) \n
    o/obj1/obj2... - Object Filter Pass all objects with the exact name of obj1, obj2, ... (* wild card allowed)\n

    """
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    if not aprs_login:
        click.echo(ctx.get_help())
        click.echo('')
        ctx.fail('Must set --aprs-login or APRS_LOGIN')
        ctx.exit()

    if not aprs_password:
        click.echo(ctx.get_help())
        click.echo('')
        ctx.fail('Must set --aprs-password or APRS_PASSWORD')
        ctx.exit()

    # CONF.aprs_network.login = aprs_login
    # config["aprs"]["password"] = aprs_password

    LOG.info(f'Python version: {sys.version}')
    LOG.info(f'APRSD Listen Started version: {aprsd.__version__}')
    utils.package.log_installed_extensions_and_plugins()

    CONF.log_opt_values(LOG, logging.DEBUG)
    collector.Collector()

    # Try and load saved MsgTrack list
    LOG.debug('Loading saved MsgTrack object.')

    # Initialize the client factory and create
    # The correct client object ready for use
    # Make sure we have 1 client transport enabled
    if not APRSDClient().is_enabled:
        LOG.error('No Clients are enabled in config.')
        sys.exit(-1)

    # Creates the client object
    LOG.info('Creating client connection')
    aprs_client = APRSDClient()
    LOG.info(aprs_client)
    if not aprs_client.login_success:
        # We failed to login, will just quit!
        msg = f'Login Failure: {aprs_client.login_failure}'
        LOG.error(msg)
        print(msg)
        sys.exit(-1)

    LOG.debug(f"Filter messages on aprsis server by '{filter}'")
    aprs_client.set_filter(filter)

    keepalive_thread = keepalive.KeepAliveThread()

    if not CONF.enable_seen_list:
        # just deregister the class from the packet collector
        packet_collector.PacketCollector().unregister(seen_list.SeenList)

    # we don't want the dupe filter to run here.
    PacketFilter().unregister(dupe_filter.DupePacketFilter)
    if packet_filter:
        LOG.info('Enabling packet filtering for {packet_filter}')
        packet_type.PacketTypeFilter().set_allow_list(packet_filter)
        PacketFilter().register(packet_type.PacketTypeFilter)
    else:
        LOG.info('No packet filtering enabled.')

    pm = None
    if load_plugins:
        pm = plugin.PluginManager()
        LOG.info('Loading plugins')
        pm.setup_plugins(load_help_plugin=False)
    elif enable_plugin:
        pm = plugin.PluginManager()
        pm.setup_plugins(
            load_help_plugin=False,
            plugin_list=enable_plugin,
        )
    else:
        LOG.warning(
            "Not Loading any plugins use --load-plugins to load what's "
            'defined in the config file.',
        )

    if pm:
        for p in pm.get_plugins():
            LOG.info('Loaded plugin %s', p.__class__.__name__)

    stats = stats_thread.APRSDStatsStoreThread()
    stats.start()

    LOG.debug('Start APRSDRxThread')
    rx_thread = rx.APRSDRXThread(packet_queue=threads.packet_queue)
    rx_thread.start()

    LOG.debug('Create APRSDListenProcessThread')
    listen_thread = APRSDListenProcessThread(
        packet_queue=threads.packet_queue,
        packet_filter=packet_filter,
        plugin_manager=pm,
        enabled_plugins=enable_plugin,
        log_packets=log_packets,
    )
    LOG.debug('Start APRSDListenProcessThread')
    listen_thread.start()
    if enable_packet_stats:
        listen_stats = ListenStatsThread()
        listen_stats.start()

    keepalive_thread.start()
    LOG.debug('keepalive Join')
    keepalive_thread.join()
    rx_thread.join()
    listen_thread.join()
    stats.join()



from aprsd import plugin
from aprsd.client import stats as client_stats
from aprsd.packets import packet_list, seen_list, tracker, watch_list
from aprsd.stats import app, collector
from aprsd.threads import aprsd


# Create the collector and register all the objects
# that APRSD has that implement the stats protocol
stats_collector = collector.Collector()
stats_collector.register_producer(app.APRSDStats)
stats_collector.register_producer(packet_list.PacketList)
stats_collector.register_producer(watch_list.WatchList)
stats_collector.register_producer(tracker.PacketTrack)
stats_collector.register_producer(plugin.PluginManager)
stats_collector.register_producer(aprsd.APRSDThreadList)
stats_collector.register_producer(client_stats.APRSClientStats)
stats_collector.register_producer(seen_list.SeenList)



import fnmatch
import importlib
import inspect
import logging
import os
import pkgutil
import sys
from traceback import print_tb

import requests
from thesmuggler import smuggle

from aprsd import plugin as aprsd_plugin

LOG = logging.getLogger()


def onerror(name):
    type, value, traceback = sys.exc_info()
    print_tb(traceback)


def plugin_type(obj):
    for c in inspect.getmro(obj):
        if issubclass(c, aprsd_plugin.APRSDRegexCommandPluginBase):
            return 'RegexCommand'
        if issubclass(c, aprsd_plugin.APRSDWatchListPluginBase):
            return 'WatchList'
        if issubclass(c, aprsd_plugin.APRSDPluginBase):
            return 'APRSDPluginBase'

    return 'Unknown'


def is_plugin(obj):
    for c in inspect.getmro(obj):
        if issubclass(c, aprsd_plugin.APRSDPluginBase):
            return True

    return False


def walk_package(package):
    return pkgutil.walk_packages(
        package.__path__,
        package.__name__ + '.',
        onerror=onerror,
    )


def get_module_info(package_name, module_name, module_path):
    if not os.path.exists(module_path):
        return None

    dir_path = os.path.realpath(module_path)
    pattern = '*.py'

    obj_list = []

    for path, _subdirs, files in os.walk(dir_path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                module = smuggle(f'{path}/{name}')
                for mem_name, obj in inspect.getmembers(module):
                    if inspect.isclass(obj) and is_plugin(obj):
                        obj_list.append(
                            {
                                'package': package_name,
                                'name': mem_name,
                                'obj': obj,
                                'version': obj.version,
                                'path': f'{".".join([module_name, obj.__name__])}',
                            },
                        )

    return obj_list


def is_aprsd_package(name):
    if name.startswith('aprsd_'):
        return True


def is_aprsd_extension(name):
    if name.startswith('aprsd_') and 'extension' in name:
        # This is an installed package that is an extension of
        # APRSD
        return True
    else:
        # We might have an editable install of an extension
        # of APRSD.
        return '__editable__' in name and 'aprsd_' in name and 'extension' in name


def get_installed_aprsd_items():
    # installed plugins
    plugins = {}
    extensions = {}
    for _finder, name, ispkg in pkgutil.iter_modules():
        if ispkg and is_aprsd_package(name):
            module = importlib.import_module(name)
            pkgs = walk_package(module)
            for pkg in pkgs:
                pkg_info = get_module_info(
                    module.__name__, pkg.name, module.__path__[0]
                )
                if 'plugin' in name:
                    plugins[name] = pkg_info
                elif 'extension' in name:
                    mod = importlib.import_module(name)
                    extensions[name] = mod
        elif is_aprsd_extension(name):
            # This isn't a package, so it could be an editable install
            module = importlib.import_module(name)
            key_name = next(iter(module.MAPPING.keys()))
            module = importlib.import_module(key_name)
            pkg_info = get_module_info(module.__name__, key_name, module.__path__[0])
            extensions[key_name] = module
    return plugins, extensions


def get_installed_plugins():
    # installed plugins
    plugins, _ = get_installed_aprsd_items()
    return plugins


def get_installed_extensions():
    # installed plugins
    _, extensions = get_installed_aprsd_items()
    return extensions


def get_pypi_packages():
    if simple_r := requests.get(
        'https://pypi.org/simple',
        headers={'Accept': 'application/vnd.pypi.simple.v1+json'},
    ):
        simple_response = simple_r.json()
    else:
        simple_response = {}

    key = 'aprsd'
    matches = [
        p['name'] for p in simple_response['projects'] if p['name'].startswith(key)
    ]

    packages = []
    for pkg in matches:
        # Get info for first match
        if r := requests.get(
            f'https://pypi.org/pypi/{pkg}/json',
            headers={'Accept': 'application/json'},
        ):
            packages.append(r.json())

    return packages


def log_installed_extensions_and_plugins():
    plugins, extensions = get_installed_aprsd_items()

    for name in extensions:
        ext = extensions[name]
        # print(f"Extension: {ext}")
        # print(f"Extension: {ext.__dict__}")
        if hasattr(ext, '__version__'):
            version = ext.__version__
        elif hasattr(ext, 'version'):
            version = ext.version
        else:
            version = ext['version']
        LOG.info(f'Extension: {name} version: {version}')

    for plugin in plugins:
        LOG.info(f'Plugin: {plugin} version: {plugins[plugin][0]["version"]}')



import unittest
from unittest import mock

from oslo_config import cfg

from aprsd import (  # noqa: F401
    conf,
    packets,
    plugins,
)
from aprsd import plugin as aprsd_plugin
from aprsd.client.drivers.registry import DriverRegistry
from aprsd.packets import core

from . import fake
from .mock_client_driver import MockClientDriver

CONF = cfg.CONF


class TestPluginManager(unittest.TestCase):
    def setUp(self) -> None:
        self.fromcall = fake.FAKE_FROM_CALLSIGN
        self.config_and_init()

        self.mock_driver = MockClientDriver()
        # Mock the DriverRegistry to return our mock driver
        self.registry_patcher = mock.patch.object(
            DriverRegistry, 'get_driver', return_value=self.mock_driver
        )
        self.mock_registry = self.registry_patcher.start()

    def tearDown(self) -> None:
        self.config = None
        aprsd_plugin.PluginManager._instance = None
        self.registry_patcher.stop()
        self.mock_registry.stop()

    def config_and_init(self):
        CONF.callsign = self.fromcall
        CONF.aprs_network.login = fake.FAKE_TO_CALLSIGN
        CONF.aprs_fi.apiKey = 'something'
        CONF.enabled_plugins = 'aprsd.plugins.ping.PingPlugin'
        CONF.enable_save = False

    def test_get_plugins_no_plugins(self):
        CONF.enabled_plugins = []
        pm = aprsd_plugin.PluginManager()
        plugin_list = pm.get_plugins()
        self.assertEqual([], plugin_list)

    def test_get_plugins_with_plugins(self):
        CONF.enabled_plugins = ['aprsd.plugins.ping.PingPlugin']
        pm = aprsd_plugin.PluginManager()
        plugin_list = pm.get_plugins()
        self.assertEqual([], plugin_list)
        pm.setup_plugins()
        plugin_list = pm.get_plugins()
        self.assertIsInstance(plugin_list, list)
        self.assertIsInstance(
            plugin_list[0],
            (
                aprsd_plugin.HelpPlugin,
                plugins.ping.PingPlugin,
            ),
        )

    def test_get_watchlist_plugins(self):
        pm = aprsd_plugin.PluginManager()
        plugin_list = pm.get_plugins()
        self.assertEqual([], plugin_list)
        pm.setup_plugins()
        plugin_list = pm.get_watchlist_plugins()
        self.assertIsInstance(plugin_list, list)
        self.assertEqual(0, len(plugin_list))

    def test_get_message_plugins(self):
        CONF.enabled_plugins = ['aprsd.plugins.ping.PingPlugin']
        pm = aprsd_plugin.PluginManager()
        plugin_list = pm.get_plugins()
        self.assertEqual([], plugin_list)
        pm.setup_plugins()
        plugin_list = pm.get_message_plugins()
        self.assertIsInstance(plugin_list, list)
        self.assertEqual(2, len(plugin_list))
        self.assertIsInstance(
            plugin_list[0],
            (
                aprsd_plugin.HelpPlugin,
                plugins.ping.PingPlugin,
            ),
        )


class TestPlugin(unittest.TestCase):
    def setUp(self) -> None:
        self.fromcall = fake.FAKE_FROM_CALLSIGN
        self.ack = 1
        self.config_and_init()

        self.mock_driver = MockClientDriver()
        # Mock the DriverRegistry to return our mock driver
        self.registry_patcher = mock.patch.object(
            DriverRegistry, 'get_driver', return_value=self.mock_driver
        )
        self.mock_registry = self.registry_patcher.start()

    def tearDown(self) -> None:
        packets.WatchList._instance = None
        packets.SeenList._instance = None
        packets.PacketTrack._instance = None
        self.config = None
        self.registry_patcher.stop()
        self.mock_registry.stop()

    def config_and_init(self):
        CONF.callsign = self.fromcall
        CONF.aprs_network.login = fake.FAKE_TO_CALLSIGN
        CONF.aprs_fi.apiKey = 'something'
        CONF.enabled_plugins = 'aprsd.plugins.ping.PingPlugin'
        CONF.enable_save = False


class TestPluginBase(TestPlugin):
    @mock.patch.object(fake.FakeBaseNoThreadsPlugin, 'process')
    def test_base_plugin_no_threads(self, mock_process):
        p = fake.FakeBaseNoThreadsPlugin()

        expected = []
        actual = p.create_threads()
        self.assertEqual(expected, actual)

        expected = '1.0'
        actual = p.version
        self.assertEqual(expected, actual)

        expected = 0
        actual = p.message_counter
        self.assertEqual(expected, actual)

        expected = None
        actual = p.filter(fake.fake_packet())
        self.assertEqual(expected, actual)
        mock_process.assert_not_called()

    @mock.patch.object(fake.FakeBaseThreadsPlugin, 'create_threads')
    def test_base_plugin_threads_created(self, mock_create):
        p = fake.FakeBaseThreadsPlugin()
        mock_create.assert_called_once()
        p.stop_threads()

    def test_base_plugin_threads(self):
        p = fake.FakeBaseThreadsPlugin()
        actual = p.create_threads()
        self.assertTrue(isinstance(actual, fake.FakeThread))
        p.stop_threads()

    @mock.patch.object(fake.FakeRegexCommandPlugin, 'process')
    def test_regex_base_not_called(self, mock_process):
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        p = fake.FakeRegexCommandPlugin()
        packet = fake.fake_packet(message='a')
        expected = None
        actual = p.filter(packet)
        self.assertEqual(expected, actual)
        mock_process.assert_not_called()

        packet = fake.fake_packet(tocall='notMe', message='f')
        expected = None
        actual = p.filter(packet)
        self.assertEqual(expected, actual)
        mock_process.assert_not_called()

        packet = fake.fake_packet(
            message_format=core.PACKET_TYPE_MICE,
        )
        expected = packets.NULL_MESSAGE
        actual = p.filter(packet)
        self.assertEqual(expected, actual)
        mock_process.assert_not_called()

        packet = fake.fake_ack_packet()
        expected = packets.NULL_MESSAGE
        actual = p.filter(packet)
        self.assertEqual(expected, actual)
        mock_process.assert_not_called()

    @mock.patch.object(fake.FakeRegexCommandPlugin, 'process')
    def test_regex_base_assert_called(self, mock_process):
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        p = fake.FakeRegexCommandPlugin()
        packet = fake.fake_packet(message='f')
        p.filter(packet)
        mock_process.assert_called_once()

    def test_regex_base_process_called(self):
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        p = fake.FakeRegexCommandPlugin()

        packet = fake.fake_packet(message='f')
        expected = fake.FAKE_MESSAGE_TEXT
        actual = p.filter(packet)
        self.assertEqual(expected, actual)

        packet = fake.fake_packet(message='F')
        expected = fake.FAKE_MESSAGE_TEXT
        actual = p.filter(packet)
        self.assertEqual(expected, actual)

        packet = fake.fake_packet(message='fake')
        expected = fake.FAKE_MESSAGE_TEXT
        actual = p.filter(packet)
        self.assertEqual(expected, actual)

        packet = fake.fake_packet(message='FAKE')
        expected = fake.FAKE_MESSAGE_TEXT
        actual = p.filter(packet)
        self.assertEqual(expected, actual)



import datetime
import tracemalloc

from oslo_config import cfg

import aprsd
from aprsd import utils
from aprsd.log import log as aprsd_log


CONF = cfg.CONF


class APRSDStats:
    """The AppStats class is used to collect stats from the application."""

    _instance = None
    start_time = None

    def __new__(cls, *args, **kwargs):
        """Have to override the new method to make this a singleton

        instead of using @singletone decorator so the unit tests work.
        """
        if not cls._instance:
            cls._instance = super().__new__(cls)
            cls._instance.start_time = datetime.datetime.now()
        return cls._instance

    def uptime(self):
        return datetime.datetime.now() - self.start_time

    def stats(self, serializable=False) -> dict:
        current, peak = tracemalloc.get_traced_memory()
        uptime = self.uptime()
        qsize = aprsd_log.logging_queue.qsize()
        if serializable:
            uptime = str(uptime)
        stats = {
            "version": aprsd.__version__,
            "uptime": uptime,
            "callsign": CONF.callsign,
            "memory_current": int(current),
            "memory_current_str": utils.human_size(current),
            "memory_peak": int(peak),
            "memory_peak_str": utils.human_size(peak),
            "loging_queue": qsize,
        }
        return stats



from unittest import mock

from oslo_config import cfg

import aprsd
from aprsd.client.drivers.fake import APRSDFakeDriver
from aprsd.plugins import version as version_plugin

from .. import fake, test_plugin

CONF = cfg.CONF


class TestVersionPlugin(test_plugin.TestPlugin):
    def setUp(self):
        # make sure the fake client driver is enabled
        # Mock CONF for testing
        super().setUp()
        self.conf_patcher = mock.patch('aprsd.client.drivers.fake.CONF')
        self.mock_conf = self.conf_patcher.start()

        # Configure fake_client.enabled
        self.mock_conf.fake_client.enabled = True

        # Create an instance of the driver
        self.driver = APRSDFakeDriver()
        self.fromcall = fake.FAKE_FROM_CALLSIGN

    def tearDown(self):
        self.conf_patcher.stop()
        super().tearDown()

    @mock.patch('aprsd.stats.collector.Collector')
    def test_version(self, mock_collector_class):
        # Set up the mock collector instance
        mock_collector_instance = mock_collector_class.return_value
        mock_collector_instance.collect.return_value = {
            'APRSDStats': {
                'uptime': '00:00:00',
            }
        }

        expected = f'APRSD ver:{aprsd.__version__} uptime:00:00:00'
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        version = version_plugin.VersionPlugin()
        version.enabled = True

        packet = fake.fake_packet(
            message='No',
            msg_number=1,
        )

        actual = version.filter(packet)
        self.assertEqual(None, actual)

        packet = fake.fake_packet(
            message='version',
            msg_number=1,
        )
        actual = version.filter(packet)
        self.assertEqual(expected, actual)

        # Verify the mock was called exactly once
        mock_collector_instance.collect.assert_called_once()



import logging
import queue
import sys

from loguru import logger
from oslo_config import cfg

from aprsd.conf import log as conf_log

CONF = cfg.CONF
# LOG = logging.getLogger("APRSD")
LOG = logger


class QueueLatest(queue.Queue):
    """Custom Queue to keep only the latest N items.

    This prevents the queue from blowing up in size.
    """

    def put(self, *args, **kwargs):
        try:
            super().put(*args, **kwargs)
        except queue.Full:
            self.queue.popleft()
            super().put(*args, **kwargs)


logging_queue = QueueLatest(maxsize=200)


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # find caller from where originated the logged message
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


# Setup the log faciility
# to disable log to stdout, but still log to file
# use the --quiet option on the cmdln
def setup_logging(loglevel=None, quiet=False, custom_handler=None):
    if not loglevel:
        log_level = CONF.logging.log_level
    else:
        log_level = conf_log.LOG_LEVELS[loglevel]

    # intercept everything at the root logger
    logging.root.handlers = [InterceptHandler()]
    logging.root.setLevel(log_level)

    # We don't really want to see the aprslib parsing debug output.
    disable_list = [
        'aprslib',
        'aprslib.parsing',
        'aprslib.exceptions',
    ]

    chardet_list = [
        'chardet',
        'chardet.charsetprober',
        'chardet.eucjpprober',
    ]

    for name in chardet_list:
        disable = logging.getLogger(name)
        disable.setLevel(logging.ERROR)

    # remove every other logger's handlers
    # and propagate to root logger
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = name not in disable_list

    handlers = []
    if CONF.logging.enable_console_stdout and not quiet:
        handlers.append(
            {
                'sink': sys.stdout,
                'serialize': False,
                'format': CONF.logging.logformat,
                'colorize': CONF.logging.enable_color,
                'level': log_level,
            },
        )

    if CONF.logging.logfile:
        handlers.append(
            {
                'sink': CONF.logging.logfile,
                'serialize': False,
                'format': CONF.logging.logformat,
                'colorize': False,
                'level': log_level,
            },
        )

    if custom_handler:
        handlers.append(custom_handler)

    # configure loguru
    logger.configure(handlers=handlers)
    logger.level('DEBUG', color='<fg #BABABA>')



import logging
from typing import Union

from oslo_config import cfg

from aprsd import packets
from aprsd.packets import core
from aprsd.utils import singleton

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


@singleton
class PacketTypeFilter:
    """This filter is used to filter out packets that don't match a specific type.

    To use this, register it with the PacketFilter class,
    then instante it and call set_allow_list() with a list of packet types
    you want to allow to pass the filtering.  All other packets will be
    filtered out.
    """

    filters = {
        packets.Packet.__name__: packets.Packet,
        packets.AckPacket.__name__: packets.AckPacket,
        packets.BeaconPacket.__name__: packets.BeaconPacket,
        packets.GPSPacket.__name__: packets.GPSPacket,
        packets.MessagePacket.__name__: packets.MessagePacket,
        packets.MicEPacket.__name__: packets.MicEPacket,
        packets.ObjectPacket.__name__: packets.ObjectPacket,
        packets.StatusPacket.__name__: packets.StatusPacket,
        packets.ThirdPartyPacket.__name__: packets.ThirdPartyPacket,
        packets.WeatherPacket.__name__: packets.WeatherPacket,
        packets.UnknownPacket.__name__: packets.UnknownPacket,
    }

    allow_list = ()

    def set_allow_list(self, filter_list):
        tmp_list = []
        for filter in filter_list:
            LOG.warning(
                f'Setting filter {filter} : {self.filters[filter]} to tmp {tmp_list}'
            )
            tmp_list.append(self.filters[filter])
        self.allow_list = tuple(tmp_list)

    def filter(self, packet: type[core.Packet]) -> Union[type[core.Packet], None]:
        """Only allow packets of certain types to filter through."""
        if self.allow_list:
            if isinstance(packet, self.allow_list):
                return packet



import threading

import wrapt
from oslo_config import cfg

from aprsd.client.client import APRSDClient
from aprsd.utils import singleton

CONF = cfg.CONF


@singleton
class APRSClientStats:
    lock = threading.Lock()

    @wrapt.synchronized(lock)
    def stats(self, serializable=False):
        return APRSDClient().stats(serializable=serializable)



import typing as t
import unittest
from unittest import mock

from click.testing import CliRunner
from oslo_config import cfg

from aprsd import conf  # noqa : F401
from aprsd.cmds import send_message  # noqa
from aprsd.main import cli

from .. import fake

CONF = cfg.CONF
F = t.TypeVar("F", bound=t.Callable[..., t.Any])


class TestSendMessageCommand(unittest.TestCase):
    def config_and_init(self, login=None, password=None):
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        CONF.trace_enabled = False
        CONF.watch_list.packet_keep_count = 1
        if login:
            CONF.aprs_network.login = login
        if password:
            CONF.aprs_network.password = password

        # CONF.aprsd_admin_extension.user = "admin"
        # CONF.aprsd_admin_extension.password = "password"

    @mock.patch("aprsd.log.log.setup_logging")
    def test_no_tocallsign(self, mock_logging):
        """Make sure we get an error if there is no tocallsign."""

        self.config_and_init(
            login="something",
            password="another",
        )
        runner = CliRunner()

        result = runner.invoke(
            cli,
            ["send-message"],
            catch_exceptions=False,
        )
        assert result.exit_code == 2
        assert "Error: Missing argument 'TOCALLSIGN'" in result.output

    @mock.patch("aprsd.log.log.setup_logging")
    def test_no_command(self, mock_logging):
        """Make sure we get an error if there is no command."""

        self.config_and_init(
            login="something",
            password="another",
        )
        runner = CliRunner()

        result = runner.invoke(
            cli,
            ["send-message", "WB4BOR"],
            catch_exceptions=False,
        )
        assert result.exit_code == 2
        assert "Error: Missing argument 'COMMAND...'" in result.output



import datetime
import decimal
import json
import sys

from aprsd.packets import core


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            args = (
                "year",
                "month",
                "day",
                "hour",
                "minute",
                "second",
                "microsecond",
            )
            return {
                "__type__": "datetime.datetime",
                "args": [getattr(obj, a) for a in args],
            }
        elif isinstance(obj, datetime.date):
            args = ("year", "month", "day")
            return {
                "__type__": "datetime.date",
                "args": [getattr(obj, a) for a in args],
            }
        elif isinstance(obj, datetime.time):
            args = ("hour", "minute", "second", "microsecond")
            return {
                "__type__": "datetime.time",
                "args": [getattr(obj, a) for a in args],
            }
        elif isinstance(obj, datetime.timedelta):
            args = ("days", "seconds", "microseconds")
            return {
                "__type__": "datetime.timedelta",
                "args": [getattr(obj, a) for a in args],
            }
        elif isinstance(obj, decimal.Decimal):
            return {
                "__type__": "decimal.Decimal",
                "args": [str(obj)],
            }
        else:
            return super().default(obj)


class SimpleJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return str(obj)
        elif isinstance(obj, datetime.time):
            return str(obj)
        elif isinstance(obj, datetime.timedelta):
            return str(obj)
        elif isinstance(obj, decimal.Decimal):
            return str(obj)
        elif isinstance(obj, core.Packet):
            return obj.to_dict()
        else:
            return super().default(obj)


class EnhancedJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            object_hook=self.object_hook,
            **kwargs,
        )

    def object_hook(self, d):
        if "__type__" not in d:
            return d
        o = sys.modules[__name__]
        for e in d["__type__"].split("."):
            o = getattr(o, e)
        args, kwargs = d.get("args", ()), d.get("kwargs", {})
        return o(*args, **kwargs)



#
# Listen on amateur radio aprs-is network for messages and respond to them.
# You must have an amateur radio callsign to use this software.  You must
# create an ~/.aprsd/config.yml file with all of the required settings.  To
# generate an example config.yml, just run aprsd, then copy the sample config
# to ~/.aprsd/config.yml and edit the settings.
#
# APRS messages:
#   l(ocation)             = descriptive location of calling station
#   w(eather)              = temp, (hi/low) forecast, later forecast
#   t(ime)                 = respond with the current time
#   f(ortune)              = respond with a short fortune
#   -email_addr email text = send an email
#   -2                     = display the last 2 emails received
#   p(ing)                 = respond with Pong!/time
#   anything else          = respond with usage
#
# (C)2018 Craig Lamparter
# License GPLv2
#

# python included libs
import datetime
import importlib.metadata as imp
import logging
import sys
import time
from importlib.metadata import version as metadata_version

import click
from oslo_config import cfg, generator

# local imports here
import aprsd
from aprsd import cli_helper, packets, threads, utils
from aprsd.stats import collector

# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
CONF = cfg.CONF
LOG = logging.getLogger('APRSD')
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(cls=cli_helper.AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option()
@click.pass_context
def cli(ctx):
    pass


def load_commands():
    from .cmds import (  # noqa
        completion,
        dev,
        fetch_stats,
        healthcheck,
        list_plugins,
        listen,
        send_message,
        server,
    )


def main():
    # First import all the possible commands for the CLI
    # The commands themselves live in the cmds directory
    load_commands()
    utils.load_entry_points('aprsd.extension')
    cli(auto_envvar_prefix='APRSD')


def signal_handler(sig, frame):
    click.echo('signal_handler: called')
    collector.Collector().stop_all()
    threads.APRSDThreadList().stop_all()
    if 'subprocess' not in str(frame):
        LOG.info(
            'Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}'.format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(1.5)
        try:
            packets.PacketTrack().save()
            packets.WatchList().save()
            packets.SeenList().save()
            packets.PacketList().save()
            collector.Collector().collect()
        except Exception as e:
            LOG.error(f'Failed to save data: {e}')
            sys.exit(0)
        # signal.signal(signal.SIGTERM, sys.exit(0))
        # sys.exit(0)


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.pass_context
@cli_helper.process_standard_options_no_config
def check_version(ctx):
    """Check this version against the latest in pypi.org."""
    level, msg = utils._check_version()
    if level:
        click.secho(msg, fg='yellow')
    else:
        click.secho(msg, fg='green')


@cli.command()
@click.pass_context
def sample_config(ctx):
    """Generate a sample Config file from aprsd and all installed plugins."""

    def _get_selected_entry_points():
        import sys

        if sys.version_info < (3, 10):
            all = imp.entry_points()
            selected = []
            if 'oslo.config.opts' in all:
                for x in all['oslo.config.opts']:
                    if x.group == 'oslo.config.opts':
                        selected.append(x)
        else:
            selected = imp.entry_points(group='oslo.config.opts')

        return selected

    def get_namespaces():
        args = []

        # selected = imp.entry_points(group="oslo.config.opts")
        selected = _get_selected_entry_points()
        for entry in selected:
            if 'aprsd' in entry.name:
                args.append('--namespace')
                args.append(entry.name)

        return args

    args = get_namespaces()
    config_version = metadata_version('oslo.config')
    logging.basicConfig(level=logging.WARN)
    conf = cfg.ConfigOpts()
    generator.register_cli_opts(conf)
    try:
        conf(args, version=config_version)
    except cfg.RequiredOptError as ex:
        conf.print_help()
        if not sys.argv[1:]:
            raise SystemExit from ex
        raise
    generator.generate(conf)
    return


@cli.command()
@click.pass_context
def version(ctx):
    """Show the APRSD version."""
    click.echo(click.style('APRSD Version : ', fg='white'), nl=False)
    click.secho(f'{aprsd.__version__}', fg='yellow', bold=True)


if __name__ == '__main__':
    main()



from unittest import mock

from oslo_config import cfg

from aprsd import conf  # noqa: F401
from aprsd.plugins import ping as ping_plugin

from .. import fake, test_plugin


CONF = cfg.CONF


class TestPingPlugin(test_plugin.TestPlugin):
    @mock.patch("time.localtime")
    def test_ping(self, mock_time):
        fake_time = mock.MagicMock()
        h = fake_time.tm_hour = 16
        m = fake_time.tm_min = 12
        s = fake_time.tm_sec = 55
        mock_time.return_value = fake_time

        CONF.callsign = fake.FAKE_TO_CALLSIGN
        ping = ping_plugin.PingPlugin()

        packet = fake.fake_packet(
            message="location",
            msg_number=1,
        )

        result = ping.filter(packet)
        self.assertEqual(None, result)

        def ping_str(h, m, s):
            return (
                "Pong! "
                + str(h).zfill(2)
                + ":"
                + str(m).zfill(2)
                + ":"
                + str(s).zfill(2)
            )

        packet = fake.fake_packet(
            message="Ping",
            msg_number=1,
        )
        actual = ping.filter(packet)
        expected = ping_str(h, m, s)
        self.assertEqual(expected, actual)

        packet = fake.fake_packet(
            message="ping",
            msg_number=1,
        )
        actual = ping.filter(packet)
        self.assertEqual(expected, actual)



from oslo_config import cfg

from aprsd.conf import client, common, log, plugin_common


CONF = cfg.CONF

log.register_opts(CONF)
common.register_opts(CONF)
client.register_opts(CONF)

# plugins
plugin_common.register_opts(CONF)


def set_lib_defaults():
    """Update default value for configuration options from other namespace.
    Example, oslo lib config options. This is needed for
    config generator tool to pick these default value changes.
    https://docs.openstack.org/oslo.config/latest/cli/
    generator.html#modifying-defaults-from-other-namespaces
    """

    # Update default value of oslo_log default_log_levels and
    # logging_context_format_string config option.
    set_log_defaults()


def set_log_defaults():
    # log.set_defaults(default_log_levels=log.get_default_log_levels())
    pass


def conf_to_dict():
    """Convert the CONF options to a single level dictionary."""
    entries = {}

    def _sanitize(opt, value):
        """Obfuscate values of options declared secret."""
        return value if not opt.secret else "*" * 4

    for opt_name in sorted(CONF._opts):
        opt = CONF._get_opt_info(opt_name)["opt"]
        val = str(_sanitize(opt, getattr(CONF, opt_name)))
        entries[str(opt)] = val

    for group_name in list(CONF._groups):
        group_attr = CONF.GroupAttr(CONF, CONF._get_group(group_name))
        for opt_name in sorted(CONF._groups[group_name]._opts):
            opt = CONF._get_opt_info(opt_name, group_name)["opt"]
            val = str(_sanitize(opt, getattr(group_attr, opt_name)))
            gname_opt_name = f"{group_name}.{opt_name}"
            entries[gname_opt_name] = val

    return entries



class RingBuffer:
    """class that implements a not-yet-full buffer"""

    max: int = 100
    data: list = []

    def __init__(self, size_max):
        self.max = size_max
        self.data = []

    class __Full:
        """class that implements a full buffer"""

        def append(self, x):
            """Append an element overwriting the oldest one."""
            self.data[self.cur] = x
            self.cur = (self.cur + 1) % self.max

        def get(self):
            """return list of elements in correct order"""
            return self.data[self.cur :] + self.data[: self.cur]

        def __len__(self):
            return len(self.data)

    def append(self, x):
        """append an element at the end of the buffer"""

        self.data.append(x)
        if len(self.data) == self.max:
            self.cur = 0
            # Permanently change self's class from non-full to full
            self.__class__ = self.__Full

    def get(self):
        """Return a list of elements from the oldest to the newest."""
        return self.data

    def __len__(self):
        return len(self.data)



from unittest import mock

from oslo_config import cfg

from aprsd import conf  # noqa: F401
from aprsd.plugins import weather as weather_plugin

from .. import fake, test_plugin

CONF = cfg.CONF


class TestUSWeatherPlugin(test_plugin.TestPlugin):
    def test_not_enabled_missing_aprs_fi_key(self):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        CONF.aprs_fi.apiKey = None
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USWeatherPlugin()
        expected = "USWeatherPlugin isn't enabled"
        packet = fake.fake_packet(message="weather")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    def test_failed_aprs_fi_location(self, mock_check):
        # When the aprs.fi api key isn't set, then
        # the Plugin will be disabled.
        mock_check.side_effect = Exception
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USWeatherPlugin()
        expected = "Failed to fetch aprs.fi location"
        packet = fake.fake_packet(message="weather")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    def test_failed_aprs_fi_location_no_entries(self, mock_check):
        # When the aprs.fi api key isn't set, then
        # the Plugin will be disabled.
        mock_check.return_value = {"entries": []}
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USWeatherPlugin()
        wx.enabled = True
        expected = "Failed to fetch aprs.fi location"
        packet = fake.fake_packet(message="weather")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    @mock.patch("aprsd.plugin_utils.get_weather_gov_for_gps")
    def test_unknown_gps(self, mock_weather, mock_check_aprs):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        mock_check_aprs.return_value = {
            "entries": [
                {
                    "lat": 10,
                    "lng": 11,
                    "lasttime": 10,
                },
            ],
        }
        mock_weather.side_effect = Exception
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USWeatherPlugin()
        wx.enabled = True
        expected = "Unable to get weather"
        packet = fake.fake_packet(message="weather")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    @mock.patch("aprsd.plugin_utils.get_weather_gov_for_gps")
    def test_working(self, mock_weather, mock_check_aprs):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        mock_check_aprs.return_value = {
            "entries": [
                {
                    "lat": 10,
                    "lng": 11,
                    "lasttime": 10,
                },
            ],
        }
        mock_weather.return_value = {
            "currentobservation": {"Temp": "400"},
            "data": {
                "temperature": ["10", "11"],
                "weather": ["test", "another"],
            },
            "time": {"startPeriodName": ["ignored", "sometime"]},
        }
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USWeatherPlugin()
        wx.enabled = True
        expected = "400F(10F/11F) test. sometime, another."
        packet = fake.fake_packet(message="weather")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)


class TestUSMetarPlugin(test_plugin.TestPlugin):
    def test_not_enabled_missing_aprs_fi_key(self):
        # When the aprs.fi api key isn't set, then
        # the LocationPlugin will be disabled.
        CONF.aprs_fi.apiKey = None
        wx = weather_plugin.USMetarPlugin()
        expected = "USMetarPlugin isn't enabled"
        packet = fake.fake_packet(message="metar")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    def test_failed_aprs_fi_location(self, mock_check):
        # When the aprs.fi api key isn't set, then
        # the Plugin will be disabled.
        mock_check.side_effect = Exception
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USMetarPlugin()
        wx.enabled = True
        expected = "Failed to fetch aprs.fi location"
        packet = fake.fake_packet(message="metar")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    def test_failed_aprs_fi_location_no_entries(self, mock_check):
        # When the aprs.fi api key isn't set, then
        # the Plugin will be disabled.
        mock_check.return_value = {"entries": []}
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USMetarPlugin()
        wx.enabled = True
        expected = "Failed to fetch aprs.fi location"
        packet = fake.fake_packet(message="metar")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_weather_gov_metar")
    def test_gov_metar_fetch_fails(self, mock_metar):
        mock_metar.side_effect = Exception
        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USMetarPlugin()
        wx.enabled = True
        expected = "Unable to find station METAR"
        packet = fake.fake_packet(message="metar KPAO")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_weather_gov_metar")
    def test_airport_works(self, mock_metar):
        class Response:
            text = '{"properties": {"rawMessage": "BOGUSMETAR"}}'

        mock_metar.return_value = Response()

        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USMetarPlugin()
        wx.enabled = True
        expected = "BOGUSMETAR"
        packet = fake.fake_packet(message="metar KPAO")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)

    @mock.patch("aprsd.plugin_utils.get_weather_gov_metar")
    @mock.patch("aprsd.plugin_utils.get_aprs_fi")
    @mock.patch("aprsd.plugin_utils.get_weather_gov_for_gps")
    def test_metar_works(self, mock_wx_for_gps, mock_check_aprs, mock_metar):
        mock_wx_for_gps.return_value = {
            "location": {"metar": "BOGUSMETAR"},
        }

        class Response:
            text = '{"properties": {"rawMessage": "BOGUSMETAR"}}'

        mock_check_aprs.return_value = {
            "entries": [
                {
                    "lat": 10,
                    "lng": 11,
                    "lasttime": 10,
                },
            ],
        }
        mock_metar.return_value = Response()

        CONF.aprs_fi.apiKey = "abc123"
        CONF.callsign = fake.FAKE_TO_CALLSIGN
        wx = weather_plugin.USMetarPlugin()
        wx.enabled = True
        expected = "BOGUSMETAR"
        packet = fake.fake_packet(message="metar")
        actual = wx.filter(packet)
        self.assertEqual(expected, actual)
